from __future__ import annotations

import logging
import logging.handlers
import subprocess
import sys
import uuid
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from typing import Any

import bitmath
import pytest
import yaml
from rich.panel import Panel
from rich.text import Text

import tuning
import tuning.logger as tuning_module
from tuning import TunedHandler, TunedLogger


def _write_yaml(path: Path, data: dict) -> None:
    with path.open("w", encoding="utf8") as file_handle:
        yaml.safe_dump(data, file_handle, sort_keys=False)


def _write_empty_defaults(path: Path) -> None:
    _write_yaml(path, {})


def _packaged_config_text() -> str:
    return Path(tuning_module.__file__).with_name("conf.yml").read_text(encoding="utf8")


def _unique_level_name(prefix: str) -> str:
    return f"{prefix}-{uuid.uuid4().hex}".upper()


def _unique_level_code() -> int:
    return 1000 + (uuid.uuid4().int % 1_000_000)


def _drop_logger(name: str) -> None:
    manager = logging.getLogger().manager
    logger = manager.loggerDict.pop(name, None)
    if isinstance(logger, logging.Logger):
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
            handler.close()


def _console_handler(logger: logging.Logger) -> TunedHandler:
    for handler in logger.handlers:
        if isinstance(handler, TunedHandler):
            return handler
    raise AssertionError("Expected a TunedHandler")


def _file_handler(logger: logging.Logger) -> Any:
    for handler in logger.handlers:
        if hasattr(handler, "baseFilename"):
            return handler
    raise AssertionError("Expected a file handler")


def _flush_handlers(logger: logging.Logger) -> None:
    for handler in logger.handlers:
        handler.flush()


def _log_record(
    *,
    level: int = logging.INFO,
    message: str = "test output",
    lineno: int = 1,
) -> logging.LogRecord:
    return logging.LogRecord(
        name="tests.console-render",
        level=level,
        pathname=__file__,
        lineno=lineno,
        msg=message,
        args=(),
        exc_info=None,
    )


def _remove_current_root_handlers(root_logger: logging.Logger) -> None:
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)


def _has_full_style_span(text: Text, style: str) -> bool:
    return any(
        span.start == 0 and span.end == len(text) and span.style == style for span in text.spans
    )


def _has_content_style_span(text: Text, style: str) -> bool:
    return any(
        span.start == 0 and span.end == len(text) - 1 and span.style == style for span in text.spans
    )


@pytest.fixture
def logger_name() -> Iterator[str]:
    name = f"tests.{uuid.uuid4().hex}"
    yield name
    _drop_logger(name)


@pytest.fixture
def clean_root_logger() -> Iterator[logging.Logger]:
    root_logger = logging.getLogger()
    original_handlers = root_logger.handlers[:]
    original_level = root_logger.level
    original_disabled = root_logger.disabled

    for handler in original_handlers:
        root_logger.removeHandler(handler)

    yield root_logger

    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)
        handler.close()

    for handler in original_handlers:
        root_logger.addHandler(handler)
    root_logger.setLevel(original_level)
    root_logger.disabled = original_disabled


def test_from_yaml_returns_tuned_logger_and_registers_levels(
    tmp_path: Path,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "app.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {"show_icon": False, "level": "CRITICAL"},
                "file": {"filename": str(log_path)},
            },
            "root": {"level": "TRACE"},
        },
    )

    logger = TunedLogger.from_yaml(config_path, name=logger_name, force=True)

    assert isinstance(logger, TunedLogger)
    assert hasattr(logger, "trace")
    assert logger.propagate is False
    assert logging.getLevelName(5) == "TRACE"

    file_handler = _file_handler(logger)
    assert file_handler.maxBytes == int(bitmath.parse_string("5 MB").bytes)
    assert log_path.parent.exists()
    assert log_path.exists()

    logger.trace("trace message")


def test_get_logger_returns_tuned_logger(logger_name: str) -> None:
    logger = tuning.getLogger(logger_name)

    assert isinstance(logger, TunedLogger)
    assert tuning.getLogger(logger_name) is logger
    assert hasattr(logger, "trace")
    assert hasattr(logger, "success")


def test_export_is_available_from_top_level_package() -> None:
    assert tuning.export is tuning_module.export


def test_add_level_is_available_from_top_level_package() -> None:
    assert tuning.addLevel is tuning_module.addLevel


def test_export_writes_packaged_config_to_directory(tmp_path: Path) -> None:
    exported_path = tuning.export(tmp_path)

    assert exported_path == (tmp_path / "tuning.yml").resolve()
    assert exported_path.read_text(encoding="utf8") == _packaged_config_text()


def test_export_writes_explicit_file_path(tmp_path: Path) -> None:
    target_path = tmp_path / "custom.yml"

    exported_path = tuning.export(target_path)

    assert exported_path == target_path.resolve()
    assert target_path.read_text(encoding="utf8") == _packaged_config_text()


def test_export_treats_nonexistent_extensionless_path_as_file(tmp_path: Path) -> None:
    target_path = tmp_path / "config"

    exported_path = tuning.export(target_path)

    assert exported_path == target_path.resolve()
    assert target_path.is_file()
    assert target_path.read_text(encoding="utf8") == _packaged_config_text()


def test_export_raises_when_target_file_exists(tmp_path: Path) -> None:
    target_path = tmp_path / "tuning.yml"
    target_path.write_text("existing config", encoding="utf8")

    with pytest.raises(FileExistsError, match="already exists"):
        tuning.export(target_path)

    assert target_path.read_text(encoding="utf8") == "existing config"


def test_export_force_overwrites_existing_file(tmp_path: Path) -> None:
    target_path = tmp_path / "tuning.yml"
    target_path.write_text("existing config", encoding="utf8")

    exported_path = tuning.export(target_path, force=True)

    assert exported_path == target_path.resolve()
    assert target_path.read_text(encoding="utf8") == _packaged_config_text()


def test_export_creates_parent_directories(tmp_path: Path) -> None:
    target_path = tmp_path / "nested" / "config" / "tuning.yml"

    exported_path = tuning.export(target_path)

    assert exported_path == target_path.resolve()
    assert target_path.read_text(encoding="utf8") == _packaged_config_text()


def test_export_none_writes_to_calling_file_directory(tmp_path: Path) -> None:
    caller_dir = tmp_path / "my-cool-app"
    caller_dir.mkdir()
    script_path = caller_dir / "main.py"
    repo_root = Path(__file__).parents[1]
    script_path.write_text(
        "import sys\n"
        f"sys.path.insert(0, {str(repo_root)!r})\n"
        "import tuning\n"
        "path = tuning.export()\n"
        "print(path)\n",
        encoding="utf8",
    )

    result = subprocess.run(
        [sys.executable, str(script_path)],
        check=True,
        capture_output=True,
        text=True,
    )

    exported_path = caller_dir / "tuning.yml"
    assert result.stdout.strip() == str(exported_path.resolve())
    assert exported_path.read_text(encoding="utf8") == _packaged_config_text()


def test_zero_config_installs_console_only_root_handler_on_first_use(
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    logger = tuning.getLogger(logger_name)
    _remove_current_root_handlers(clean_root_logger)

    logger.info("zero config output")

    assert len(clean_root_logger.handlers) == 1
    assert isinstance(clean_root_logger.handlers[0], TunedHandler)
    assert not any(
        isinstance(handler, logging.FileHandler) for handler in clean_root_logger.handlers
    )
    assert clean_root_logger.level == logging.INFO


def test_zero_config_dynamic_level_methods_log_from_default_metadata(
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    logger = tuning.getLogger(logger_name)
    logger_as_any: Any = logger
    _remove_current_root_handlers(clean_root_logger)

    logger_as_any.success("zero config success")

    assert len(clean_root_logger.handlers) == 1
    assert isinstance(clean_root_logger.handlers[0], TunedHandler)


def test_zero_config_does_not_replace_existing_root_handlers(
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    _remove_current_root_handlers(clean_root_logger)
    existing_handler = logging.StreamHandler()
    clean_root_logger.addHandler(existing_handler)
    logger = tuning.getLogger(logger_name)

    logger.info("handled elsewhere")

    assert clean_root_logger.handlers == [existing_handler]


def test_zero_config_prompt_uses_default_prompt_style(
    clean_root_logger: logging.Logger,
    logger_name: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    captured: dict[str, Any] = {}
    logger = tuning.getLogger(logger_name)
    _remove_current_root_handlers(clean_root_logger)

    def fake_input(self, prompt, **kwargs):
        captured["prompt"] = prompt
        captured["kwargs"] = kwargs
        return "zero prompt value"

    monkeypatch.setattr(tuning_module.Console, "input", fake_input)

    value = logger.prompt("Zero prompt")

    assert value == "zero prompt value"
    assert len(clean_root_logger.handlers) == 1
    assert isinstance(clean_root_logger.handlers[0], TunedHandler)
    assert "<<<" in captured["prompt"].plain
    assert "Zero prompt" in captured["prompt"].plain
    assert _has_content_style_span(captured["prompt"], "italic bold black on magenta")


def test_basic_config_configures_root_for_child_loggers(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "myapp.log"
    logger = tuning.getLogger(logger_name)

    tuning.basicConfig(filename=log_path, level="INFO", force=True)

    logger.info("Started")
    logger.debug("Hidden")
    _flush_handlers(clean_root_logger)

    content = log_path.read_text(encoding="utf8")
    assert "Started" in content
    assert "Hidden" not in content
    assert logger.handlers == []
    assert len(clean_root_logger.handlers) == 1


def test_basic_config_is_noop_when_root_already_has_handlers(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "ignored.log"

    tuning.basicConfig(level="INFO", force=True)
    first_handlers = clean_root_logger.handlers[:]
    tuning.basicConfig(filename=log_path, level="ERROR")

    assert clean_root_logger.handlers == first_handlers
    assert isinstance(clean_root_logger.handlers[0], TunedHandler)
    assert not log_path.exists()


def test_basic_config_applies_console_handler_options(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(
        level="INFO",
        force=True,
        show_icon=True,
        show_level=False,
        show_path=True,
        show_time=True,
        boxes=True,
        rich_tracebacks=False,
        markup=False,
    )

    handler = _console_handler(clean_root_logger)

    assert handler.show_icon is True
    assert handler._log_render.show_level is False
    assert handler._log_render.show_path is True
    assert handler._log_render.show_time is True
    assert handler.boxes is True
    assert handler.rich_tracebacks is False
    assert handler.markup is False


def test_basic_config_show_level_false_renders_message_only(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True, show_level=False)
    handler = _console_handler(clean_root_logger)
    record = _log_record(message="message only output")
    message = handler.render_message(record, record.getMessage())

    rendered = handler.render(record=record, traceback=None, message_renderable=message)

    assert handler._log_render.show_level is False
    assert len(rendered.columns) == 1
    body: Any = rendered.columns[0]._cells[0]
    assert body._renderables == [message]


def test_iso_format_is_exported_for_console_timestamps() -> None:
    assert tuning.ISO_FORMAT == "[%Y-%m-%d %H:%M:%S]"


def test_rotation_defaults_are_exported() -> None:
    assert tuning.DEFAULT_MAX_BYTES == int(bitmath.parse_string("10 MB").bytes)
    assert tuning.DEFAULT_BACKUP_COUNT == 3


def test_basic_config_datefmt_controls_console_timestamp(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(
        level="INFO",
        force=True,
        show_time=True,
        datefmt=tuning.ISO_FORMAT,
    )
    handler = _console_handler(clean_root_logger)
    record = _log_record()
    record.created = datetime(2026, 5, 4, 1, 58, 41).timestamp()

    rendered_time = handler._render_log_time(record)

    assert rendered_time.plain == "[2026-05-04 01:58:41] "


def test_yaml_log_time_format_controls_console_timestamp(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {
                    "()": "tuning.TunedHandler",
                    "level": "INFO",
                    "show_time": True,
                    "log_time_format": tuning.ISO_FORMAT,
                }
            },
            "root": {"level": "INFO", "handlers": ["console"]},
        },
    )

    tuning.basicConfigFromYaml(
        config_path,
        defaults_path=defaults_path,
        force=True,
    )
    handler = _console_handler(clean_root_logger)
    record = _log_record()
    record.created = datetime(2026, 5, 4, 1, 58, 41).timestamp()

    rendered_time = handler._render_log_time(record)

    assert rendered_time.plain == "[2026-05-04 01:58:41] "


def test_console_handler_applies_level_style_to_message(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True)
    handler = _console_handler(clean_root_logger)
    record = logging.LogRecord(
        name="tests.console-style",
        level=logging.INFO,
        pathname=__file__,
        lineno=1,
        msg="styled output",
        args=(),
        exc_info=None,
    )

    message = handler.render_message(record, record.getMessage())

    assert isinstance(message, Text)
    assert message.plain == "styled output"
    assert _has_full_style_span(message, "cyan")


def test_console_handler_uses_compact_symbol_prefix_width(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True)
    handler = _console_handler(clean_root_logger)
    record = logging.LogRecord(
        name="tests.console-prefix",
        level=logging.INFO,
        pathname=__file__,
        lineno=1,
        msg="compact output",
        args=(),
        exc_info=None,
    )

    level_text = handler.get_level_text(record)

    assert level_text.plain == "[*]"
    assert level_text.cell_len == 3
    assert _has_full_style_span(level_text, "cyan")


def test_console_handler_styles_level_separator_with_prefix(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True)
    handler = _console_handler(clean_root_logger)
    record = logging.LogRecord(
        name="tests.console-prefix",
        level=logging.INFO,
        pathname=__file__,
        lineno=1,
        msg="compact output",
        args=(),
        exc_info=None,
    )

    level_text = handler._render_level_with_separator(record)

    assert level_text.plain == "[*] "
    assert level_text.cell_len == 4
    assert _has_full_style_span(level_text, "cyan")


def test_console_handler_uses_compact_icon_prefix_width(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True, show_icon=True)
    handler = _console_handler(clean_root_logger)
    record = logging.LogRecord(
        name="tests.console-prefix",
        level=logging.INFO,
        pathname=__file__,
        lineno=1,
        msg="compact output",
        args=(),
        exc_info=None,
    )

    level_text = handler.get_level_text(record)

    assert level_text.plain.startswith("🔵")
    assert level_text.cell_len == 3
    assert _has_full_style_span(level_text, "cyan")


def test_console_handler_styles_background_level_separator(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="CRITICAL", force=True)
    handler = _console_handler(clean_root_logger)
    record = logging.LogRecord(
        name="tests.console-prefix",
        level=logging.CRITICAL,
        pathname=__file__,
        lineno=1,
        msg="critical output",
        args=(),
        exc_info=None,
    )

    level_text = handler._render_level_with_separator(record)

    assert level_text.plain == "!!! "
    assert _has_full_style_span(level_text, "bold white on red")


def test_console_handler_renders_boxed_record_with_level_title(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="INFO", force=True, boxes=True)
    handler = _console_handler(clean_root_logger)
    record = _log_record(message="boxed output")
    message = handler.render_message(record, record.getMessage())

    rendered = handler.render(
        record=record,
        traceback="traceback output",
        message_renderable=message,
    )

    panel = rendered.columns[0]._cells[0]
    assert isinstance(panel, Panel)
    assert isinstance(panel.title, Text)
    assert panel.title.plain == "[*] INFO"
    assert panel.border_style == "cyan"
    assert panel.style == "cyan"
    assert _has_full_style_span(panel.title, "cyan")
    panel_body: Any = panel.renderable
    assert panel_body._renderables == [message, "traceback output"]


def test_boxed_render_styles_critical_panel_fill(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(level="CRITICAL", force=True, boxes=True)
    handler = _console_handler(clean_root_logger)
    record = _log_record(level=logging.CRITICAL, message="critical boxed output")
    message = handler.render_message(record, record.getMessage())

    rendered = handler.render(record=record, traceback=None, message_renderable=message)

    panel = rendered.columns[0]._cells[0]
    assert isinstance(panel, Panel)
    assert isinstance(panel.title, Text)
    assert panel.title.plain == "!!! CRITICAL"
    assert panel.border_style == "bold white on red"
    assert panel.style == "bold white on red"
    assert _has_full_style_span(panel.title, "bold white on red")


def test_boxed_render_keeps_time_and_path_outside_panel(
    clean_root_logger: logging.Logger,
) -> None:
    tuning.basicConfig(
        level="INFO",
        force=True,
        boxes=True,
        show_time=True,
        show_path=True,
    )
    handler = _console_handler(clean_root_logger)
    record = _log_record(message="boxed output", lineno=123)
    message = handler.render_message(record, record.getMessage())

    rendered = handler.render(record=record, traceback=None, message_renderable=message)

    assert len(rendered.columns) == 3
    assert isinstance(rendered.columns[1]._cells[0], Panel)
    assert rendered.columns[0]._cells[0].plain.endswith(" ")
    assert "test_tuned_logger.py:123" in rendered.columns[2]._cells[0].plain


def test_boxed_render_honors_hidden_level() -> None:
    handler = TunedHandler(boxes=True, show_level=False, show_time=False)
    record = _log_record(message="untitled boxed output")
    message = handler.render_message(record, record.getMessage())

    rendered = handler.render(record=record, traceback=None, message_renderable=message)

    panel = rendered.columns[0]._cells[0]
    assert isinstance(panel, Panel)
    assert panel.title is None


def test_console_handler_keeps_long_fallback_labels_untruncated() -> None:
    handler = TunedHandler()
    record = logging.LogRecord(
        name="tests.console-prefix",
        level=123456,
        pathname=__file__,
        lineno=1,
        msg="long label output",
        args=(),
        exc_info=None,
    )
    record.levelname = "MY-LONG-LEVEL"

    level_text = handler.get_level_text(record)

    assert level_text.plain == "[MY-LONG-LEVEL]"
    assert level_text.cell_len > 3
    assert _has_full_style_span(level_text, "logging.level.my-long-level")


def test_basic_config_force_replaces_root_handlers(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "forced.log"
    logger = tuning.getLogger(logger_name)

    tuning.basicConfig(level="INFO")
    tuning.basicConfig(filename=log_path, level="INFO", force=True)

    logger.info("forced output")
    _flush_handlers(clean_root_logger)

    assert len(clean_root_logger.handlers) == 1
    assert "forced output" in log_path.read_text(encoding="utf8")


def test_basic_config_filename_is_file_only_and_uses_detailed_formatter(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "file-only.log"
    logger = tuning.getLogger(logger_name)

    tuning.basicConfig(
        filename=log_path,
        level="INFO",
        force=True,
        show_icon=True,
        show_path=True,
        show_time=True,
        rich_tracebacks=False,
        markup=False,
    )

    logger.info("file-only output")
    _flush_handlers(clean_root_logger)

    assert len(clean_root_logger.handlers) == 1
    assert not isinstance(clean_root_logger.handlers[0], TunedHandler)
    assert clean_root_logger.handlers[0].formatter is not None
    assert (
        clean_root_logger.handlers[0].formatter._fmt
        == "%(asctime)s [%(levelname)s] %(name)s %(funcName)s: %(message)s"
    )
    content = log_path.read_text(encoding="utf8")
    assert "[INFO]" in content
    assert logger_name in content
    assert "test_basic_config_filename_is_file_only_and_uses_detailed_formatter" in content
    assert "file-only output" in content


def test_basic_config_filename_with_console_creates_file_and_console_handlers(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "with-console.log"
    logger = tuning.getLogger(logger_name)

    tuning.basicConfig(
        filename=log_path,
        console=True,
        level="INFO",
        force=True,
        show_icon=True,
    )

    logger.info("console and file output")
    _flush_handlers(clean_root_logger)

    assert len(clean_root_logger.handlers) == 2
    assert isinstance(clean_root_logger.handlers[0], logging.FileHandler)
    assert _console_handler(clean_root_logger).show_icon is True
    assert "console and file output" in log_path.read_text(encoding="utf8")


def test_basic_config_filename_without_rotation_uses_plain_file_handler(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "plain.log"

    tuning.basicConfig(filename=log_path, level="INFO", force=True)

    file_handler = clean_root_logger.handlers[0]
    assert isinstance(file_handler, logging.FileHandler)
    assert not isinstance(file_handler, logging.handlers.RotatingFileHandler)


def test_basic_config_max_bytes_uses_rotating_file_handler(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "rotating.log"

    tuning.basicConfig(
        filename=log_path,
        level="INFO",
        force=True,
        max_bytes=1024,
        backup_count=5,
    )

    file_handler = clean_root_logger.handlers[0]
    assert isinstance(file_handler, logging.handlers.RotatingFileHandler)
    assert file_handler.maxBytes == 1024
    assert file_handler.backupCount == 5
    assert file_handler.formatter is not None
    assert (
        file_handler.formatter._fmt
        == "%(asctime)s [%(levelname)s] %(name)s %(funcName)s: %(message)s"
    )


def test_basic_config_accepts_human_readable_max_bytes(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "human-readable.log"

    tuning.basicConfig(
        filename=log_path,
        level="INFO",
        force=True,
        max_bytes="2 MB",
        backup_count=2,
    )

    file_handler = clean_root_logger.handlers[0]
    assert isinstance(file_handler, logging.handlers.RotatingFileHandler)
    assert file_handler.maxBytes == int(bitmath.parse_string("2 MB").bytes)
    assert file_handler.backupCount == 2


def test_basic_config_max_bytes_without_backup_count_warns_and_uses_default(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "default-backups.log"

    with pytest.warns(UserWarning, match="DEFAULT_BACKUP_COUNT"):
        tuning.basicConfig(
            filename=log_path,
            level="INFO",
            force=True,
            max_bytes="1 MB",
        )

    file_handler = clean_root_logger.handlers[0]
    assert isinstance(file_handler, logging.handlers.RotatingFileHandler)
    assert file_handler.maxBytes == int(bitmath.parse_string("1 MB").bytes)
    assert file_handler.backupCount == tuning.DEFAULT_BACKUP_COUNT


def test_basic_config_backup_count_without_max_bytes_warns_and_uses_default(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
) -> None:
    log_path = tmp_path / "logs" / "default-max-bytes.log"

    with pytest.warns(UserWarning, match="DEFAULT_MAX_BYTES"):
        tuning.basicConfig(
            filename=log_path,
            level="INFO",
            force=True,
            backup_count=4,
        )

    file_handler = clean_root_logger.handlers[0]
    assert isinstance(file_handler, logging.handlers.RotatingFileHandler)
    assert file_handler.maxBytes == tuning.DEFAULT_MAX_BYTES
    assert file_handler.backupCount == 4


def test_basic_config_ignores_rotation_options_without_filename(
    clean_root_logger: logging.Logger,
) -> None:
    kwargs: dict[str, Any] = {
        "level": "INFO",
        "force": True,
        "max_bytes": object(),
        "backup_count": -1,
    }

    tuning.basicConfig(**kwargs)

    assert len(clean_root_logger.handlers) == 1
    assert isinstance(clean_root_logger.handlers[0], TunedHandler)


@pytest.mark.parametrize("max_bytes", [0, -1, True, object(), "not a size"])
def test_basic_config_invalid_max_bytes_raises(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    max_bytes: Any,
) -> None:
    log_path = tmp_path / "logs" / "invalid-size.log"

    with pytest.raises(ValueError, match="max_bytes"):
        tuning.basicConfig(
            filename=log_path,
            level="INFO",
            force=True,
            max_bytes=max_bytes,
            backup_count=1,
        )


@pytest.mark.parametrize("backup_count", [-1, True, object(), "3"])
def test_basic_config_invalid_backup_count_raises(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    backup_count: Any,
) -> None:
    log_path = tmp_path / "logs" / "invalid-backups.log"

    with pytest.raises(ValueError, match="backup_count"):
        tuning.basicConfig(
            filename=log_path,
            level="INFO",
            force=True,
            max_bytes=1024,
            backup_count=backup_count,
        )


@pytest.mark.parametrize(
    "option",
    [
        "console",
        "show_icon",
        "show_level",
        "show_path",
        "show_time",
        "rich_tracebacks",
        "markup",
    ],
)
def test_basic_config_boolean_options_must_be_boolean(
    clean_root_logger: logging.Logger,
    option: str,
) -> None:
    kwargs: dict[str, Any] = {option: "yes"}

    with pytest.raises(ValueError, match=f"{option} must be a boolean"):
        tuning.basicConfig(force=True, **kwargs)


def test_basic_config_installs_default_custom_level_methods(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "trace.log"
    logger = tuning.getLogger(logger_name)

    tuning.basicConfig(filename=log_path, level="TRACE", force=True)

    logger_as_any: Any = logger
    logger_as_any.trace("trace output")
    _flush_handlers(clean_root_logger)

    assert "trace output" in log_path.read_text(encoding="utf8")


def test_basic_config_from_yaml_configures_actual_root_logger(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "root.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "root": {"level": "INFO", "handlers": ["file"]},
        },
    )
    logger = tuning.getLogger(logger_name)

    tuning.basicConfigFromYaml(config_path, defaults_path=defaults_path, force=True)

    logger.info("root inherited output")
    _flush_handlers(clean_root_logger)

    assert logger.handlers == []
    assert len(clean_root_logger.handlers) == 1
    assert "root inherited output" in log_path.read_text(encoding="utf8")


def test_basic_config_from_yaml_without_path_loads_packaged_defaults(
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    logger = tuning.getLogger(logger_name)

    tuning.basicConfigFromYaml(force=True)

    logger.info("packaged default output")
    _flush_handlers(clean_root_logger)

    assert any(isinstance(handler, TunedHandler) for handler in clean_root_logger.handlers)
    assert any(isinstance(handler, logging.FileHandler) for handler in clean_root_logger.handlers)


def test_basic_config_from_yaml_prompt_uses_inherited_root_handler(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    captured: dict[str, Any] = {}
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "prompt": {"symbol": "??", "icon": "!!", "style": "green"},
            "handlers": {
                "console": {
                    "()": "tuning.TunedHandler",
                    "level": "DEBUG",
                    "show_icon": True,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["console"]},
        },
    )
    logger = tuning.getLogger(logger_name)

    tuning.basicConfigFromYaml(config_path, defaults_path=defaults_path, force=True)
    handler = _console_handler(clean_root_logger)

    def fake_input(prompt, **kwargs):
        captured["prompt"] = prompt
        captured["kwargs"] = kwargs
        return "typed root value"

    monkeypatch.setattr(handler.console, "input", fake_input)

    value = logger.prompt("Root prompt")

    assert value == "typed root value"
    assert "!!" in captured["prompt"].plain
    assert "Root prompt" in captured["prompt"].plain
    assert captured["prompt"].plain.endswith(" ")
    assert _has_content_style_span(captured["prompt"], "green")
    assert not _has_full_style_span(captured["prompt"], "green")


def test_prompt_uses_the_first_handler_show_icon_setting(
    tmp_path: Path,
    logger_name: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    log_path = tmp_path / "logs" / "app.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {"show_icon": True},
                "file": {"filename": str(log_path)},
            }
        },
    )

    logger = TunedLogger.from_yaml(config_path, name=logger_name, force=True)
    handler = _console_handler(logger)
    captured: dict[str, Any] = {}

    def fake_input(prompt, **kwargs):
        captured["prompt"] = prompt
        captured["kwargs"] = kwargs
        return "typed value"

    monkeypatch.setattr(handler.console, "input", fake_input)

    value = logger.prompt("Enter a value")

    assert value == "typed value"
    assert "📝" in captured["prompt"].plain
    assert "Enter a value" in captured["prompt"].plain
    assert captured["prompt"].plain.endswith(" ")
    assert _has_content_style_span(captured["prompt"], "italic bold black on magenta")
    assert not _has_full_style_span(captured["prompt"], "italic bold black on magenta")
    assert captured["kwargs"]["markup"] is False
    assert captured["kwargs"]["password"] is False


def test_invalid_builtin_level_code_raises_a_clear_error(
    tmp_path: Path,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "app.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "levels": {
                "DEBUG": {"code": 20},
            },
            "handlers": {
                "file": {"filename": str(log_path)},
            },
        },
    )

    with pytest.raises(ValueError, match="Built-in level DEBUG"):
        TunedLogger.from_yaml(config_path, name=logger_name, force=True)


def test_from_yaml_is_idempotent_and_force_reconfigures(
    tmp_path: Path,
    logger_name: str,
) -> None:
    log_path_one = tmp_path / "logs-one" / "app.log"
    log_path_two = tmp_path / "logs-two" / "app.log"
    config_path_one = tmp_path / "logger-one.yml"
    config_path_two = tmp_path / "logger-two.yml"
    _write_yaml(
        config_path_one,
        {
            "handlers": {
                "console": {"show_icon": False},
                "file": {"filename": str(log_path_one)},
            }
        },
    )
    _write_yaml(
        config_path_two,
        {
            "handlers": {
                "console": {"show_icon": True},
                "file": {"filename": str(log_path_two)},
            }
        },
    )

    logger = TunedLogger.from_yaml(config_path_one, name=logger_name, force=True)
    same_logger = TunedLogger.from_yaml(config_path_one, name=logger_name)
    assert logger is same_logger

    with pytest.raises(ValueError, match="different config"):
        TunedLogger.from_yaml(config_path_two, name=logger_name)

    reconfigured_logger = TunedLogger.from_yaml(
        config_path_two,
        name=logger_name,
        force=True,
    )
    assert logger is reconfigured_logger
    assert _console_handler(reconfigured_logger).show_icon is True
    assert Path(_file_handler(reconfigured_logger).baseFilename) == log_path_two.resolve()


def test_existing_standard_logger_name_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "app.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "handlers": {
                "file": {"filename": str(log_path)},
            }
        },
    )

    standard_logger = logging.getLogger(logger_name)
    assert not isinstance(standard_logger, TunedLogger)

    with pytest.raises(ValueError, match="expected TunedLogger"):
        TunedLogger.from_yaml(config_path, name=logger_name, force=True)


def test_partial_override_deep_merges_with_defaults(
    tmp_path: Path,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "merged.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {
                    "level": "ERROR",
                    "show_icon": True,
                },
                "file": {"filename": str(log_path)},
            },
            "root": {"level": "ERROR"},
        },
    )

    logger = TunedLogger.from_yaml(config_path, name=logger_name, force=True)

    console_handler = _console_handler(logger)
    file_handler = _file_handler(logger)
    assert logger.level == logging.ERROR
    assert console_handler.level == logging.ERROR
    assert console_handler.show_icon is True
    assert console_handler.boxes is False
    assert console_handler.markup is True
    assert file_handler.maxBytes == int(bitmath.parse_string("5 MB").bytes)
    assert file_handler.backupCount == 3


def test_boxes_from_yaml_configures_tuned_handler(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {
                    "()": "tuning.TunedHandler",
                    "boxes": True,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["console"]},
        },
    )

    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )

    assert _console_handler(logger).boxes is True


def test_custom_level_conflict_by_name_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path_one = tmp_path / "logger-one.yml"
    config_path_two = tmp_path / "logger-two.yml"
    log_path_one = tmp_path / "logs" / "one.log"
    log_path_two = tmp_path / "logs" / "two.log"
    level_name = _unique_level_name("name-conflict")
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path_one,
        {
            "levels": {level_name: {"code": _unique_level_code()}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path_one)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )
    _write_yaml(
        config_path_two,
        {
            "levels": {level_name: {"code": _unique_level_code()}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path_two)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    TunedLogger.from_yaml(
        config_path_one, name=logger_name, defaults_path=defaults_path, force=True
    )

    with pytest.raises(ValueError, match="different definition"):
        TunedLogger.from_yaml(
            config_path_two, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_custom_level_conflict_by_code_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path_one = tmp_path / "logger-one.yml"
    config_path_two = tmp_path / "logger-two.yml"
    log_path_one = tmp_path / "logs" / "one.log"
    log_path_two = tmp_path / "logs" / "two.log"
    level_code = _unique_level_code()
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path_one,
        {
            "levels": {_unique_level_name("code-conflict-one"): {"code": level_code}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path_one)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )
    _write_yaml(
        config_path_two,
        {
            "levels": {_unique_level_name("code-conflict-two"): {"code": level_code}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path_two)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    TunedLogger.from_yaml(
        config_path_one, name=logger_name, defaults_path=defaults_path, force=True
    )

    with pytest.raises(ValueError, match="already registered"):
        TunedLogger.from_yaml(
            config_path_two, name=logger_name, defaults_path=defaults_path, force=True
        )


@pytest.mark.parametrize("alias, canonical", [("WARN", "WARNING"), ("FATAL", "CRITICAL")])
def test_level_aliases_are_rejected(
    tmp_path: Path,
    logger_name: str,
    alias: str,
    canonical: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "levels": {alias: {"code": 30}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    with pytest.raises(ValueError, match=f"Use {canonical} instead of {alias}"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_input_level_is_rejected_in_favor_of_prompt(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "levels": {"INPUT": {"code": 15}},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    with pytest.raises(ValueError, match="top-level prompt"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_show_icon_on_non_custom_rich_handler_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "stream": {
                    "class": "logging.StreamHandler",
                    "show_icon": True,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["stream"]},
        },
    )

    with pytest.raises(ValueError, match="not a TunedHandler"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_show_icon_must_be_boolean(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {
                    "()": "tuning.TunedHandler",
                    "show_icon": "yes",
                }
            },
            "root": {"level": "DEBUG", "handlers": ["console"]},
        },
    )

    with pytest.raises(ValueError, match="show_icon must be a boolean"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_boxes_on_non_custom_rich_handler_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "stream": {
                    "class": "logging.StreamHandler",
                    "boxes": True,
                }
            },
            "root": {"level": "DEBUG", "handlers": ["stream"]},
        },
    )

    with pytest.raises(ValueError, match="not a TunedHandler"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_boxes_must_be_boolean(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {
                    "()": "tuning.TunedHandler",
                    "boxes": "yes",
                }
            },
            "root": {"level": "DEBUG", "handlers": ["console"]},
        },
    )

    with pytest.raises(ValueError, match="boxes must be a boolean"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_invalid_human_readable_max_bytes_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {
                "file": {
                    "class": "logging.handlers.RotatingFileHandler",
                    "filename": str(log_path),
                    "maxBytes": "not a size",
                }
            },
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    with pytest.raises(ValueError, match="Invalid maxBytes"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_explicit_requested_logger_config_is_supported(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "loggers": {
                logger_name: {
                    "level": "INFO",
                    "handlers": ["file"],
                    "propagate": False,
                }
            },
        },
    )

    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )

    assert logger.level == logging.INFO
    assert logger.propagate is False
    assert log_path.exists()


def test_unexpected_logger_entries_raise(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "loggers": {
                "other.logger": {
                    "level": "INFO",
                    "handlers": ["file"],
                }
            },
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    with pytest.raises(ValueError, match="unexpected entries"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_config_without_root_or_requested_logger_raises(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
        },
    )

    with pytest.raises(ValueError, match="either root or the requested logger"):
        TunedLogger.from_yaml(
            config_path, name=logger_name, defaults_path=defaults_path, force=True
        )


def test_prompt_uses_symbol_when_icons_are_disabled(
    tmp_path: Path,
    logger_name: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    log_path = tmp_path / "logs" / "app.log"
    config_path = tmp_path / "logger.yml"
    _write_yaml(
        config_path,
        {
            "handlers": {
                "console": {"show_icon": False},
                "file": {"filename": str(log_path)},
            }
        },
    )

    logger = TunedLogger.from_yaml(config_path, name=logger_name, force=True)
    handler = _console_handler(logger)
    captured: dict[str, Any] = {}

    def fake_input(prompt, **kwargs):
        captured["prompt"] = prompt
        captured["kwargs"] = kwargs
        return "typed value"

    monkeypatch.setattr(handler.console, "input", fake_input)

    value = logger.prompt("Enter a value")

    assert value == "typed value"
    assert "<<<" in captured["prompt"].plain
    assert "Enter a value" in captured["prompt"].plain


def test_prompt_falls_back_to_new_console_without_custom_rich_handler(
    tmp_path: Path,
    logger_name: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "app.log"
    captured: dict[str, Any] = {}
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "prompt": {"symbol": "??", "icon": "!!", "style": "green"},
            "handlers": {"file": {"class": "logging.FileHandler", "filename": str(log_path)}},
            "root": {"level": "DEBUG", "handlers": ["file"]},
        },
    )

    def fake_input(self, prompt, **kwargs):
        captured["prompt"] = prompt
        captured["kwargs"] = kwargs
        return "fallback value"

    monkeypatch.setattr(tuning_module.Console, "input", fake_input)
    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )

    value = logger.prompt("Fallback prompt")

    assert value == "fallback value"
    assert "??" in captured["prompt"].plain
    assert "Fallback prompt" in captured["prompt"].plain
    assert captured["kwargs"]["markup"] is False
    assert captured["kwargs"]["password"] is False


def test_add_level_registers_runtime_method_for_existing_and_future_loggers(
    tmp_path: Path,
    clean_root_logger: logging.Logger,
    logger_name: str,
) -> None:
    log_path = tmp_path / "logs" / "runtime.log"
    level_name = _unique_level_name("runtime-level")
    level_code = _unique_level_code()
    method_name = level_name.lower().replace("-", "_")
    logger = tuning.getLogger(logger_name)
    future_logger_name = f"{logger_name}.future"

    with pytest.raises(AttributeError):
        getattr(logger, method_name)("before registration")

    try:
        tuning.addLevel(level_code, level_name, symbol="RL", style="green")
        future_logger = tuning.getLogger(future_logger_name)
        tuning.basicConfig(filename=log_path, level=level_name, force=True)

        getattr(logger, method_name)("existing runtime output")
        getattr(future_logger, method_name)("future runtime output")
        _flush_handlers(clean_root_logger)
    finally:
        _drop_logger(future_logger_name)

    content = log_path.read_text(encoding="utf8")
    assert "existing runtime output" in content
    assert "future runtime output" in content


def test_add_level_registers_console_metadata(
    clean_root_logger: logging.Logger,
) -> None:
    level_name = _unique_level_name("metadata-level")
    level_code = _unique_level_code()
    tuning.addLevel(level_code, level_name, symbol="MT", icon="RUN", style="bold green")
    tuning.basicConfig(level=level_name, force=True, show_icon=True)
    handler = _console_handler(clean_root_logger)
    record = _log_record(level=level_code)

    level_text = handler.get_level_text(record)
    message = handler.render_message(record, record.getMessage())

    assert level_text.plain == "RUN"
    assert _has_full_style_span(level_text, "bold green")
    assert isinstance(message, Text)
    assert _has_full_style_span(message, "bold green")


def test_add_level_is_idempotent_for_same_definition() -> None:
    level_name = _unique_level_name("idempotent-level")
    level_code = _unique_level_code()

    tuning.addLevel(level_code, level_name, symbol="ID", style="blue")
    tuning.addLevel(level_code, level_name, symbol="ID", style="blue")

    assert logging.getLevelName(level_code) == level_name


def test_add_level_conflicting_name_raises() -> None:
    level_name = _unique_level_name("name-conflict")
    tuning.addLevel(_unique_level_code(), level_name, symbol="NC")

    with pytest.raises(ValueError, match="different definition"):
        tuning.addLevel(_unique_level_code(), level_name, symbol="NC")


def test_add_level_conflicting_code_raises() -> None:
    level_code = _unique_level_code()
    tuning.addLevel(level_code, _unique_level_name("code-conflict-one"), symbol="C1")

    with pytest.raises(ValueError, match="already registered|reserved"):
        tuning.addLevel(level_code, _unique_level_name("code-conflict-two"), symbol="C2")


def test_add_level_method_conflict_raises_without_partial_registration() -> None:
    level_prefix = f"METHOD-CONFLICT-{uuid.uuid4().hex}"
    first_name = f"{level_prefix}-ONE"
    second_name = f"{level_prefix}_ONE"
    second_code = _unique_level_code()
    tuning.addLevel(_unique_level_code(), first_name, symbol="M1")

    with pytest.raises(ValueError, match="already bound"):
        tuning.addLevel(second_code, second_name, symbol="M2")

    assert logging.getLevelName(second_code) == f"Level {second_code}"


@pytest.mark.parametrize(
    "name",
    ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "WARN", "FATAL", "INPUT"],
)
def test_add_level_rejects_reserved_names(name: str) -> None:
    with pytest.raises(ValueError):
        tuning.addLevel(_unique_level_code(), name)


@pytest.mark.parametrize("name", ["", "MY CUSTOM", "1_LEVEL", "MY.LEVEL", "class"])
def test_add_level_rejects_invalid_names(name: str) -> None:
    with pytest.raises(ValueError, match="name"):
        tuning.addLevel(_unique_level_code(), name)


@pytest.mark.parametrize("num", [True, object(), "7"])
def test_add_level_rejects_invalid_level_codes(num: Any) -> None:
    with pytest.raises(ValueError, match="num must be an integer"):
        tuning.addLevel(num, _unique_level_name("invalid-code"))


@pytest.mark.parametrize(
    "kwargs, message",
    [
        ({"symbol": "TOO-LONG"}, "symbol"),
        ({"icon": object()}, "icon"),
    ],
)
def test_add_level_rejects_invalid_metadata(
    kwargs: dict[str, Any],
    message: str,
) -> None:
    with pytest.raises(ValueError, match=message):
        tuning.addLevel(_unique_level_code(), _unique_level_name("invalid-meta"), **kwargs)


def test_custom_level_with_hyphen_creates_snake_case_method_and_logs(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "custom.log"
    level_name = _unique_level_name("method-level")
    level_code = _unique_level_code()
    method_name = level_name.lower().replace("-", "_")
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "levels": {
                level_name: {
                    "code": level_code,
                    "symbol": "ML",
                    "style": "green",
                }
            },
            "handlers": {
                "file": {
                    "class": "logging.FileHandler",
                    "filename": str(log_path),
                    "level": level_name,
                }
            },
            "root": {"level": level_name, "handlers": ["file"]},
        },
    )

    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )
    log_method = getattr(logger, method_name)

    log_method("custom method message")
    for handler in logger.handlers:
        handler.flush()

    assert "custom method message" in log_path.read_text(encoding="utf8")


def test_custom_level_reports_user_call_site(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "caller.log"
    level_name = _unique_level_name("caller-level")
    level_code = _unique_level_code()
    method_name = level_name.lower().replace("-", "_")
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "levels": {level_name: {"code": level_code, "symbol": "CL"}},
            "formatters": {"caller": {"format": "%(filename)s:%(funcName)s:%(message)s"}},
            "handlers": {
                "file": {
                    "class": "logging.FileHandler",
                    "filename": str(log_path),
                    "formatter": "caller",
                    "level": level_name,
                }
            },
            "root": {"level": level_name, "handlers": ["file"]},
        },
    )

    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )
    log_method = getattr(logger, method_name)

    log_method("custom caller message")
    for handler in logger.handlers:
        handler.flush()

    content = log_path.read_text(encoding="utf8")
    assert "test_tuned_logger.py:test_custom_level_reports_user_call_site" in content
    assert "_levels.py" not in content


def test_custom_level_respects_user_stacklevel(
    tmp_path: Path,
    logger_name: str,
) -> None:
    defaults_path = tmp_path / "defaults.yml"
    config_path = tmp_path / "logger.yml"
    log_path = tmp_path / "logs" / "stacklevel.log"
    level_name = _unique_level_name("stacklevel")
    level_code = _unique_level_code()
    method_name = level_name.lower().replace("-", "_")
    _write_empty_defaults(defaults_path)
    _write_yaml(
        config_path,
        {
            "levels": {level_name: {"code": level_code, "symbol": "SL"}},
            "formatters": {"caller": {"format": "%(filename)s:%(funcName)s:%(message)s"}},
            "handlers": {
                "file": {
                    "class": "logging.FileHandler",
                    "filename": str(log_path),
                    "formatter": "caller",
                    "level": level_name,
                }
            },
            "root": {"level": level_name, "handlers": ["file"]},
        },
    )

    logger = TunedLogger.from_yaml(
        config_path, name=logger_name, defaults_path=defaults_path, force=True
    )
    log_method = getattr(logger, method_name)

    def helper() -> None:
        log_method("custom stacklevel message", stacklevel=2)

    helper()
    for handler in logger.handlers:
        handler.flush()

    content = log_path.read_text(encoding="utf8")
    assert "test_tuned_logger.py:test_custom_level_respects_user_stacklevel" in content
    assert "helper" not in content
    assert "_levels.py" not in content



from __future__ import annotations

import keyword
import logging
import re
from typing import Any

from rich.style import Style

from tuning._models import LevelSpec

_BUILTIN_LEVELS = {
    "NOTSET": logging.NOTSET,
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}
_ALIASED_LEVEL_NAMES = {
    "WARN": "WARNING",
    "FATAL": "CRITICAL",
}
_LEVEL_SPECS_BY_NAME: dict[str, LevelSpec] = {}
_LEVEL_SPECS_BY_CODE: dict[int, LevelSpec] = {}
_DYNAMIC_METHODS: dict[str, int] = {}
_RUNTIME_LEVEL_NAME_PATTERN = re.compile(r"^[A-Z][A-Z0-9_-]*$")


def get_level_spec(level_name: str) -> LevelSpec | None:
    return _LEVEL_SPECS_BY_NAME.get(level_name)


def parse_level_specs(levels_config: dict[str, Any]) -> list[LevelSpec]:
    if not isinstance(levels_config, dict):
        raise ValueError("levels must be a mapping")

    specs: list[LevelSpec] = []
    seen_names: set[str] = set()
    seen_codes: set[int] = set()

    for raw_name, raw_spec in levels_config.items():
        if not isinstance(raw_spec, dict):
            raise ValueError(f"Level {raw_name} must be configured as a mapping")

        name = _normalize_level_name(raw_name)
        if name == "INPUT":
            raise ValueError("Use the top-level prompt section instead of levels.INPUT")

        if name in _ALIASED_LEVEL_NAMES:
            canonical_name = _ALIASED_LEVEL_NAMES[name]
            raise ValueError(f"Use {canonical_name} instead of {name} in levels")

        code = _coerce_level_code(name, raw_spec.get("code"))
        expected_builtin_code = _BUILTIN_LEVELS.get(name)
        if expected_builtin_code is not None and code != expected_builtin_code:
            raise ValueError(
                f"Built-in level {name} must use stdlib code {expected_builtin_code}, not {code}"
            )

        if name in seen_names:
            raise ValueError(f"Duplicate level name: {name}")
        if code in seen_codes:
            raise ValueError(f"Duplicate level code: {code}")

        symbol = validate_symbol(raw_spec.get("symbol"), field_name=f"levels.{raw_name}.symbol")
        icon = raw_spec.get("icon")
        if icon is not None and not isinstance(icon, str):
            raise ValueError(f"levels.{raw_name}.icon must be a string when provided")

        style = validate_style(raw_spec.get("style"), field_name=f"levels.{raw_name}.style")

        seen_names.add(name)
        seen_codes.add(code)
        specs.append(LevelSpec(name=name, code=code, symbol=symbol, icon=icon, style=style))

    return specs


def parse_runtime_level_spec(
    *,
    num: int,
    name: str,
    symbol: str | None,
    icon: str | None,
    style: str | None,
) -> LevelSpec:
    if isinstance(num, bool) or not isinstance(num, int):
        raise ValueError("num must be an integer")

    if not isinstance(name, str) or not name.strip():
        raise ValueError("name must be a non-empty string")

    normalized_name = _normalize_level_name(name)
    if normalized_name == "INPUT":
        raise ValueError("Use prompts instead of registering an INPUT level")

    if normalized_name in _ALIASED_LEVEL_NAMES:
        canonical_name = _ALIASED_LEVEL_NAMES[normalized_name]
        raise ValueError(f"Use {canonical_name} instead of {normalized_name}")

    if normalized_name in _BUILTIN_LEVELS:
        raise ValueError(f"Cannot add built-in level {normalized_name}")

    if not _RUNTIME_LEVEL_NAME_PATTERN.fullmatch(normalized_name):
        raise ValueError(
            "name must start with a letter and contain only letters, numbers, '_' or '-'"
        )

    method_name = _normalize_method_name(normalized_name)
    if not method_name.isidentifier() or keyword.iskeyword(method_name):
        raise ValueError("name must create a valid Python method name")

    if icon is not None and not isinstance(icon, str):
        raise ValueError("icon must be a string when provided")

    return LevelSpec(
        name=normalized_name,
        code=num,
        symbol=validate_symbol(symbol, field_name="symbol"),
        icon=icon,
        style=validate_style(style, field_name="style"),
    )


def register_level_specs(specs: list[LevelSpec]) -> None:
    for spec in specs:
        _register_level_spec(spec)


def validate_level_specs(specs: list[LevelSpec]) -> None:
    for spec in specs:
        _validate_level_spec(spec)


def validate_dynamic_level_methods(
    logger_cls: type[logging.Logger], specs: list[LevelSpec]
) -> None:
    for spec in specs:
        if spec.name in _BUILTIN_LEVELS:
            continue

        method_name = _normalize_method_name(spec.name)
        _validate_dynamic_level_method(logger_cls, method_name, spec.code)


def install_dynamic_level_methods(logger_cls: type[logging.Logger], specs: list[LevelSpec]) -> None:
    for spec in specs:
        if spec.name in _BUILTIN_LEVELS:
            continue

        method_name = _normalize_method_name(spec.name)
        if _validate_dynamic_level_method(logger_cls, method_name, spec.code):
            continue

        setattr(logger_cls, method_name, _make_level_method(spec.code, spec.name, method_name))
        _DYNAMIC_METHODS[method_name] = spec.code


def validate_symbol(symbol: str | None, *, field_name: str) -> str | None:
    if symbol is None:
        return None

    if not isinstance(symbol, str) or not symbol:
        raise ValueError(f"{field_name} must be a non-empty string when provided")

    if len(symbol) > 3:
        raise ValueError(f"{field_name} must be at most 3 characters")

    return symbol


def validate_style(style: str | None, *, field_name: str) -> str | None:
    if style is None:
        return None

    if not isinstance(style, str):
        raise ValueError(f"{field_name} must be a string when provided")

    Style.parse(style)
    return style


def _normalize_level_name(raw_name: str) -> str:
    return raw_name.strip().upper()


def _normalize_method_name(level_name: str) -> str:
    return level_name.lower().replace("-", "_")


def _coerce_level_code(level_name: str, code: Any) -> int:
    if isinstance(code, bool):
        raise ValueError(f"Level {level_name} code must be an integer")

    if isinstance(code, int):
        return code

    if isinstance(code, str) and code.strip().isdigit():
        return int(code.strip())

    raise ValueError(f"Level {level_name} code must be an integer")


def _register_level_spec(spec: LevelSpec) -> None:
    _validate_level_spec(spec)

    if spec.name not in _BUILTIN_LEVELS:
        logging.addLevelName(spec.code, spec.name)

    _LEVEL_SPECS_BY_NAME[spec.name] = spec
    _LEVEL_SPECS_BY_CODE[spec.code] = spec


def _validate_level_spec(spec: LevelSpec) -> None:
    existing_by_name = _LEVEL_SPECS_BY_NAME.get(spec.name)
    if existing_by_name and existing_by_name != spec:
        raise ValueError(f"Level {spec.name} is already registered with a different definition")

    existing_by_code = _LEVEL_SPECS_BY_CODE.get(spec.code)
    if existing_by_code and existing_by_code != spec:
        raise ValueError(
            f"Level code {spec.code} is already registered for {existing_by_code.name}"
        )

    known_levels = logging.getLevelNamesMapping()
    if spec.name in known_levels and known_levels[spec.name] != spec.code:
        raise ValueError(
            f"Level {spec.name} is already registered with code {known_levels[spec.name]}"
        )

    known_name = logging.getLevelName(spec.code)
    if (
        isinstance(known_name, str)
        and not known_name.startswith("Level ")
        and known_name != spec.name
    ):
        raise ValueError(f"Level code {spec.code} is already reserved for {known_name}")


def _validate_dynamic_level_method(
    logger_cls: type[logging.Logger], method_name: str, level_code: int
) -> bool:
    existing_level = _DYNAMIC_METHODS.get(method_name)
    if existing_level is not None:
        if existing_level != level_code:
            raise ValueError(
                f"Method {method_name} is already bound to level {existing_level}, not {level_code}"
            )
        return True

    if hasattr(logger_cls, method_name):
        raise ValueError(f"Cannot create dynamic method {method_name}: name already exists")

    return False


def _make_level_method(level_code: int, level_name: str, method_name: str):
    def log_for_level(self: logging.Logger, msg: str, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(level_code):
            kwargs["stacklevel"] = kwargs.get("stacklevel", 1) + 1
            self._log(level_code, msg, args, **kwargs)

    log_for_level.__name__ = method_name
    log_for_level.__qualname__ = f"TunedLogger.{method_name}"
    log_for_level.__doc__ = f"Log a message with the {level_name} level."
    return log_for_level



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class LevelSpec:
    """Metadata attached to a logging level.

    Attributes:
        name: Level name, for example `INFO` or `SUCCESS`.
        code: Numeric stdlib logging level code.
        symbol: Compact text prefix used when icons are disabled.
        icon: Icon prefix used when `show_icon=True`.
        style: Rich style applied to console level prefixes and messages.
    """

    name: str
    code: int
    symbol: str | None = None
    icon: str | None = None
    style: str | None = None


@dataclass(frozen=True)
class PromptSpec:
    """Metadata used to render styled interactive prompts.

    Attributes:
        symbol: Prompt prefix used when icons are disabled.
        icon: Prompt prefix used when icons are enabled.
        style: Rich style applied to the prompt prefix and question text.
    """

    symbol: str | None = None
    icon: str | None = None
    style: str | None = None



from tuning._banners import banner
from tuning.logger import (
    DEFAULT_BACKUP_COUNT,
    DEFAULT_MAX_BYTES,
    ISO_FORMAT,
    LevelSpec,
    PromptSpec,
    TunedHandler,
    TunedLogger,
    addLevel,
    basicConfig,
    basicConfigFromYaml,
    export,
    getLogger,
)

__all__ = [
    "TunedLogger",
    "TunedHandler",
    "LevelSpec",
    "PromptSpec",
    "ISO_FORMAT",
    "DEFAULT_MAX_BYTES",
    "DEFAULT_BACKUP_COUNT",
    "getLogger",
    "banner",
    "export",
    "addLevel",
    "basicConfig",
    "basicConfigFromYaml",
]



from __future__ import annotations

from typing import Any

from rich.text import Text

from tuning._levels import validate_style, validate_symbol
from tuning._models import PromptSpec


def parse_prompt_spec(prompt_config: dict[str, Any]) -> PromptSpec:
    if not isinstance(prompt_config, dict):
        raise ValueError("prompt must be a mapping")

    symbol = validate_symbol(prompt_config.get("symbol"), field_name="prompt.symbol")
    icon = prompt_config.get("icon")
    if icon is not None and not isinstance(icon, str):
        raise ValueError("prompt.icon must be a string when provided")

    style = validate_style(prompt_config.get("style"), field_name="prompt.style")
    return PromptSpec(symbol=symbol, icon=icon, style=style)


def render_prompt_prefix(prompt_spec: PromptSpec, *, show_icon: bool) -> Text:
    prefix = _select_prompt_prefix(prompt_spec, show_icon=show_icon)
    style = prompt_spec.style or ""
    if style:
        return Text.styled(prefix, style)

    return Text(prefix)


def render_prompt_text(
    prompt_spec: PromptSpec,
    message: str,
    *,
    show_icon: bool,
    markup: bool,
) -> Text:
    prefix = _select_prompt_prefix(prompt_spec, show_icon=show_icon)
    message_text = Text.from_markup(message) if markup else Text(message)
    prompt_text = Text.assemble(prefix, " ", message_text)

    if prompt_spec.style:
        prompt_text.stylize(prompt_spec.style, 0, len(prompt_text))

    prompt_text.append(" ")
    return prompt_text


def _select_prompt_prefix(prompt_spec: PromptSpec, *, show_icon: bool) -> str:
    if show_icon and prompt_spec.icon:
        return prompt_spec.icon
    elif prompt_spec.symbol:
        return prompt_spec.symbol

    return ">"



from __future__ import annotations

import json
from copy import deepcopy
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any

import bitmath
import yaml

from tuning._levels import parse_level_specs
from tuning._models import LevelSpec, PromptSpec
from tuning._prompt import parse_prompt_spec

_DEFAULT_CONFIG_PATH = Path(__file__).resolve().parent / "conf.yml"


@dataclass(frozen=True)
class ResolvedConfig:
    logging_config: dict[str, Any]
    level_specs: list[LevelSpec]
    prompt_spec: PromptSpec
    signature: str


@dataclass(frozen=True)
class TuningMetadata:
    level_specs: list[LevelSpec]
    prompt_spec: PromptSpec


def load_tuning_config(
    config_path: str | Path,
    *,
    logger_name: str,
    defaults_path: str | Path | None = None,
) -> ResolvedConfig:
    base_config = _load_yaml(defaults_path or _DEFAULT_CONFIG_PATH)
    override_config = _load_yaml(config_path)
    merged_config = _deep_merge(base_config, override_config)

    level_specs = parse_level_specs(merged_config.get("levels", {}))
    prompt_spec = parse_prompt_spec(merged_config.get("prompt", {}))
    normalized_config = _normalize_named_logging_config(merged_config, logger_name)
    signature = _build_signature(normalized_config, level_specs, prompt_spec)

    return ResolvedConfig(
        logging_config=normalized_config,
        level_specs=level_specs,
        prompt_spec=prompt_spec,
        signature=signature,
    )


def load_tuning_root_config(
    config_path: str | Path | None = None,
    *,
    defaults_path: str | Path | None = None,
) -> ResolvedConfig:
    base_config = _load_yaml(defaults_path or _DEFAULT_CONFIG_PATH)
    if config_path is None:
        merged_config = base_config
    else:
        override_config = _load_yaml(config_path)
        merged_config = _deep_merge(base_config, override_config)

    level_specs = parse_level_specs(merged_config.get("levels", {}))
    prompt_spec = parse_prompt_spec(merged_config.get("prompt", {}))
    normalized_config = _normalize_root_logging_config(merged_config)
    signature = _build_signature(normalized_config, level_specs, prompt_spec)

    return ResolvedConfig(
        logging_config=normalized_config,
        level_specs=level_specs,
        prompt_spec=prompt_spec,
        signature=signature,
    )


def load_tuning_metadata(*, defaults_path: str | Path | None = None) -> TuningMetadata:
    config = _load_yaml(defaults_path or _DEFAULT_CONFIG_PATH)
    return TuningMetadata(
        level_specs=parse_level_specs(config.get("levels", {})),
        prompt_spec=parse_prompt_spec(config.get("prompt", {})),
    )


def export_default_config(path: str | Path, *, force: bool = False) -> Path:
    target_path = Path(path).expanduser()
    if target_path.exists() and not force:
        raise FileExistsError(f"Config file already exists: {target_path}")

    target_path.parent.mkdir(parents=True, exist_ok=True)
    target_path.write_text(_DEFAULT_CONFIG_PATH.read_text(encoding="utf8"), encoding="utf8")
    return target_path.resolve()


def _load_yaml(path_value: str | Path) -> dict[str, Any]:
    path = Path(path_value)
    with path.open("r", encoding="utf8") as file_handle:
        data = yaml.safe_load(file_handle) or {}

    if not isinstance(data, dict):
        raise ValueError(f"Expected a mapping in YAML config: {path}")

    return data


def _deep_merge(base: Any, override: Any) -> Any:
    if isinstance(base, dict) and isinstance(override, dict):
        merged = deepcopy(base)
        for key, value in override.items():
            if key in merged:
                merged[key] = _deep_merge(merged[key], value)
            else:
                merged[key] = deepcopy(value)
        return merged

    return deepcopy(override)


def _normalize_named_logging_config(config: dict[str, Any], logger_name: str) -> dict[str, Any]:
    normalized = deepcopy(config)
    normalized.pop("levels", None)
    normalized.pop("prompt", None)

    _normalize_common_logging_config(normalized)

    logger_config = _select_logger_config(normalized, logger_name)
    normalized.pop("root", None)
    normalized["loggers"] = {logger_name: logger_config}

    return normalized


def _normalize_root_logging_config(config: dict[str, Any]) -> dict[str, Any]:
    normalized = deepcopy(config)
    normalized.pop("levels", None)
    normalized.pop("prompt", None)

    _normalize_common_logging_config(normalized)

    loggers = normalized.get("loggers")
    if loggers is not None and not isinstance(loggers, dict):
        raise ValueError("loggers must be a mapping when provided")

    root_config = normalized.get("root")
    if root_config is None:
        raise ValueError("Config must define root for basicConfigFromYaml")

    if not isinstance(root_config, dict):
        raise ValueError("Root configuration must be a mapping")

    return normalized


def _normalize_common_logging_config(normalized: dict[str, Any]) -> None:
    version = normalized.get("version", 1)
    if version != 1:
        raise ValueError(f"Unsupported logging config version: {version}")

    normalized["version"] = 1
    normalized["disable_existing_loggers"] = False

    handlers = normalized.get("handlers", {})
    if not isinstance(handlers, dict) or not handlers:
        raise ValueError("Config must define at least one handler")

    for handler_name, handler_config in handlers.items():
        if not isinstance(handler_config, dict):
            raise ValueError(f"Handler {handler_name} must be configured as a mapping")
        _normalize_handler_config(handler_name, handler_config)


def _normalize_handler_config(handler_name: str, handler_config: dict[str, Any]) -> None:
    for tuning_option in ("show_icon", "boxes"):
        if tuning_option not in handler_config:
            continue

        if not _is_custom_rich_handler(handler_config):
            raise ValueError(
                f"Handler {handler_name} uses {tuning_option} but is not a TunedHandler"
            )

        if not isinstance(handler_config[tuning_option], bool):
            raise ValueError(f"handlers.{handler_name}.{tuning_option} must be a boolean")

    if "maxBytes" in handler_config and isinstance(handler_config["maxBytes"], str):
        handler_config["maxBytes"] = parse_size_to_bytes(
            handler_config["maxBytes"], field_name="maxBytes"
        )

    filename = handler_config.get("filename")
    if filename is not None:
        Path(filename).expanduser().parent.mkdir(parents=True, exist_ok=True)


def _is_custom_rich_handler(handler_config: dict[str, Any]) -> bool:
    factory = handler_config.get("()") or handler_config.get("class")
    if isinstance(factory, str):
        return factory in {
            "tuning.logger.TunedHandler",
            "tuning.TunedHandler",
        }

    return (
        getattr(factory, "__module__", None) == "tuning.logger"
        and getattr(factory, "__qualname__", None) == "TunedHandler"
    )


def parse_size_to_bytes(raw_value: str, *, field_name: str) -> int:
    try:
        size = bitmath.parse_string(raw_value)
    except ValueError as error:
        raise ValueError(f"Invalid {field_name} value: {raw_value}") from error

    return int(size.bytes)


def _select_logger_config(config: dict[str, Any], logger_name: str) -> dict[str, Any]:
    loggers = config.get("loggers")
    root_config = config.get("root")

    if loggers is not None and not isinstance(loggers, dict):
        raise ValueError("loggers must be a mapping when provided")

    if loggers:
        extra_loggers = [name for name in loggers if name != logger_name]
        if extra_loggers:
            raise ValueError(
                "v1 only supports configuring the requested logger name; "
                f"unexpected entries: {', '.join(extra_loggers)}"
            )

    if loggers and logger_name in loggers:
        logger_config = deepcopy(loggers[logger_name])
    elif root_config is not None:
        logger_config = deepcopy(root_config)
    else:
        raise ValueError("Config must define either root or the requested logger entry")

    if not isinstance(logger_config, dict):
        raise ValueError("Logger configuration must be a mapping")

    logger_config.setdefault("propagate", False)
    return logger_config


def _build_signature(
    normalized_config: dict[str, Any],
    level_specs: list[LevelSpec],
    prompt_spec: PromptSpec,
) -> str:
    payload = {
        "config": normalized_config,
        "levels": [asdict(spec) for spec in level_specs],
        "prompt": asdict(prompt_spec),
    }
    return json.dumps(payload, sort_keys=True, separators=(",", ":"), ensure_ascii=True)



from __future__ import annotations

import inspect
import random
import re
from pathlib import Path
from typing import TypeAlias

from rich import box as rich_box
from rich.box import Box
from rich.cells import cell_len
from rich.console import Console
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

_BANNER_FILENAME = "banners.txt"
_BANNER_NAME_PATTERN = re.compile(r"^[a-z0-9]+(?:_[a-z0-9]+)*$")
_FALLBACK_MESSAGE = "We could display a beautiful banner here, but the terminal is too tight :("
_RANDOM_STYLE_COLORS = (
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
)
MEGA_BOLD = Box("████\n█  █\n████\n█  █\n████\n█  █\n████\n████\n")
_PaddingSpec: TypeAlias = tuple[int] | tuple[int, int] | tuple[int, int, int, int]
_BoxSpec: TypeAlias = str | Box


def banner(
    path: str | Path | None = None,
    name: str | None = None,
    *,
    border_style: str | None = None,
    text_style: str | None = None,
    padding: tuple[int] | tuple[int, int] | tuple[int, int, int, int] = (0, 0),
    border: bool = True,
    box: str | Box | None = None,
    background_style: str | None = None,
) -> str | None:
    """Print an application banner from a banner file.

    Banner files contain one or more sections marked by `### banner_name`.
    When `path` is omitted, `banners.txt` is discovered by walking upward from
    the calling file's directory. If the caller has no file, discovery starts
    from the current working directory.

    Args:
        path: Banner file path, directory containing `banners.txt`, or `None`
            to discover `banners.txt` from the caller location.
        name: Optional banner name. If omitted, a random banner that fits the
            current terminal width is selected.
        border_style: Rich style for the panel border. If omitted, a random
            bold color is used.
        text_style: Rich style for the banner text. If omitted, a random bold
            color is used.
        padding: Rich-compatible padding tuple. Accepts `(all,)`,
            `(vertical, horizontal)`, or `(top, right, bottom, left)`.
        border: Render the banner inside a Rich panel when true. When false,
            only styled text and padding are rendered.
        box: Optional Rich box style name or `rich.box.Box` object. String
            names are matched case-insensitively against `rich.box` constants
            and the custom `MEGA_BOLD` style.
        background_style: Optional Rich style for the panel background. When
            `border=False`, this style is applied to the padding area.

    Returns:
        The exact banner body printed, or `None` when no banner fits the
        terminal and the fallback message is printed instead.

    Raises:
        ValueError: If no banner file can be discovered, or if the banner file
            is malformed.
        FileNotFoundError: If an explicit banner file path does not exist.
        KeyError: If `name` is provided but no matching banner exists.
    """
    resolved_padding = _normalize_padding(padding)
    if not isinstance(border, bool):
        raise ValueError("border must be a boolean")
    resolved_box = _resolve_box(box)
    resolved_border_style = border_style or _random_style()
    resolved_text_style = text_style or _random_style()
    banner_path = _resolve_banner_path(path)
    banners = _parse_banners(banner_path.read_text(encoding="utf8"), banner_path)
    console = Console()
    selected_banner = _select_banner(banners, name, console.width, resolved_padding, border)

    if selected_banner is None:
        _print_banner(
            _FALLBACK_MESSAGE,
            console,
            border_style=resolved_border_style,
            text_style=resolved_text_style,
            padding=padding,
            border=border,
            box=resolved_box,
            background_style=background_style,
        )
        return None

    _print_banner(
        selected_banner,
        console,
        border_style=resolved_border_style,
        text_style=resolved_text_style,
        padding=padding,
        border=border,
        box=resolved_box,
        background_style=background_style,
    )
    return selected_banner


def _random_style() -> str:
    return f"bold {random.choice(_RANDOM_STYLE_COLORS)}"


def _resolve_banner_path(path: str | Path | None) -> Path:
    if path is None:
        return _discover_banner_path()

    resolved_path = Path(path).expanduser()
    if resolved_path.is_dir():
        resolved_path = resolved_path / _BANNER_FILENAME
    return resolved_path


def _discover_banner_path() -> Path:
    start_dir = _caller_directory() or Path.cwd()
    for directory in (start_dir, *start_dir.parents):
        candidate = directory / _BANNER_FILENAME
        if candidate.is_file():
            return candidate

    raise ValueError("Banner file not defined")


def _caller_directory() -> Path | None:
    current_file = Path(__file__).resolve()
    frame = inspect.currentframe()
    while frame is not None:
        filename = frame.f_code.co_filename
        if filename and not filename.startswith("<"):
            candidate = Path(filename).resolve()
            if candidate != current_file:
                return candidate.parent
        frame = frame.f_back

    return None


def _parse_banners(content: str, source: Path) -> dict[str, str]:
    banners: dict[str, str] = {}
    current_name: str | None = None
    current_lines: list[str] = []
    lines = content.splitlines(keepends=True)

    for index, line in enumerate(lines):
        if not line.startswith("###"):
            if current_name is not None:
                current_lines.append(line)
            elif line.strip():
                raise ValueError(f"Banner content must start with a marker: {source}")
            continue

        current_name = _store_banner(banners, current_name, current_lines, source)
        current_lines = []
        marker_name = line[3:].strip()
        if not marker_name:
            if any(remaining_line.strip() for remaining_line in lines[index + 1 :]):
                raise ValueError(f"Empty banner marker must terminate the file: {source}")
            current_name = None
            break
        if _BANNER_NAME_PATTERN.fullmatch(marker_name) is None:
            raise ValueError(f"Invalid banner name {marker_name!r}: {source}")
        current_name = marker_name

    _store_banner(banners, current_name, current_lines, source)
    if not banners:
        raise ValueError(f"No banners defined: {source}")

    return banners


def _store_banner(
    banners: dict[str, str],
    name: str | None,
    lines: list[str],
    source: Path,
) -> str | None:
    if name is None:
        return None
    if name in banners:
        raise ValueError(f"Duplicate banner name {name!r}: {source}")

    banner_text = "".join(lines)
    if not banner_text:
        raise ValueError(f"Banner {name!r} has no content: {source}")

    banners[name] = banner_text
    return None


def _select_banner(
    banners: dict[str, str],
    name: str | None,
    console_width: int,
    padding: tuple[int, int, int, int],
    border: bool,
) -> str | None:
    if name is not None:
        if name not in banners:
            raise KeyError(name)
        selected_banner = banners[name]
        if _banner_fits(selected_banner, console_width, padding, border):
            return selected_banner
        return None

    if len(banners) == 1:
        selected_banner = next(iter(banners.values()))
        if _banner_fits(selected_banner, console_width, padding, border):
            return selected_banner
        return None

    fitting_banners = [
        banner_text
        for banner_text in banners.values()
        if _banner_fits(banner_text, console_width, padding, border)
    ]
    if not fitting_banners:
        return None

    return random.choice(fitting_banners)


def _banner_fits(
    banner_text: str,
    console_width: int,
    padding: tuple[int, int, int, int],
    border: bool,
) -> bool:
    return _banner_width(banner_text) + _horizontal_overhead(padding, border) <= console_width


def _banner_width(banner_text: str) -> int:
    return max((cell_len(line.rstrip("\r\n")) for line in banner_text.splitlines()), default=0)


def _horizontal_overhead(padding: tuple[int, int, int, int], border: bool) -> int:
    _, right, _, left = padding
    border_width = 2 if border else 0
    return border_width + right + left


def _normalize_padding(padding: _PaddingSpec) -> tuple[int, int, int, int]:
    if not isinstance(padding, tuple):
        raise ValueError("padding must be a tuple of 1, 2, or 4 non-negative integers")
    if len(padding) not in {1, 2, 4}:
        raise ValueError("padding must be a tuple of 1, 2, or 4 non-negative integers")
    if any(not isinstance(value, int) or value < 0 for value in padding):
        raise ValueError("padding must be a tuple of 1, 2, or 4 non-negative integers")

    if len(padding) == 1:
        top = right = bottom = left = padding[0]
    elif len(padding) == 2:
        top = bottom = padding[0]
        right = left = padding[1]
    else:
        top, right, bottom, left = padding

    return top, right, bottom, left


def _resolve_box(box: _BoxSpec | None) -> Box | None:
    if box is None:
        return None
    if isinstance(box, Box):
        return box
    if not isinstance(box, str):
        raise ValueError("box must be a rich.box style name, MEGA_BOLD, or rich.box.Box")

    box_name = box.strip().upper()
    if box_name == "MEGA_BOLD":
        return MEGA_BOLD

    rich_box_style = getattr(rich_box, box_name, None)
    if isinstance(rich_box_style, Box):
        return rich_box_style

    raise ValueError("box must be one of the styles defined in rich.box or MEGA_BOLD")


def _print_banner(
    banner_text: str,
    console: Console,
    *,
    border_style: str,
    text_style: str,
    padding: _PaddingSpec,
    border: bool,
    box: Box | None,
    background_style: str | None,
) -> None:
    text = Text(banner_text, style=text_style)
    if not border:
        if background_style is None:
            console.print(Padding(text, padding))
        else:
            console.print(Padding(text, padding, style=background_style))
        return

    if box is None and background_style is None:
        panel = Panel.fit(text, border_style=border_style, padding=padding)
    elif box is None:
        assert background_style is not None
        panel = Panel.fit(
            text,
            border_style=border_style,
            padding=padding,
            style=background_style,
        )
    elif background_style is None:
        panel = Panel.fit(text, box=box, border_style=border_style, padding=padding)
    else:
        assert background_style is not None
        panel = Panel.fit(
            text,
            box=box,
            border_style=border_style,
            padding=padding,
            style=background_style,
        )

    console.print(panel)



from __future__ import annotations

import subprocess
import sys
from collections.abc import Sequence
from pathlib import Path

import pytest
from rich import box as rich_box
from rich.box import Box
from rich.console import Console
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

import tuning
import tuning._banners as banners_module


def _write_banners(path: Path, content: str) -> None:
    path.write_text(content, encoding="utf8")


def _set_console_width(monkeypatch: pytest.MonkeyPatch, width: int) -> None:
    def console_factory() -> Console:
        return Console(width=width, color_system=None)

    monkeypatch.setattr(banners_module, "Console", console_factory)


class _RecordingConsole:
    def __init__(self, width: int = 80) -> None:
        self.width = width
        self.renderables: list[object] = []

    def print(self, renderable: object) -> None:
        self.renderables.append(renderable)


def _record_console(monkeypatch: pytest.MonkeyPatch, width: int = 80) -> _RecordingConsole:
    console = _RecordingConsole(width=width)
    monkeypatch.setattr(banners_module, "Console", lambda: console)
    return console


def test_banner_is_available_from_top_level_package() -> None:
    assert tuning.banner is banners_module.banner


def test_banner_prints_named_banner_from_explicit_file_path(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    _set_console_width(monkeypatch, 80)
    banner_path = tmp_path / "my-banners.txt"
    expected = "\nTUNING\n"
    _write_banners(
        banner_path,
        "### first\nnot selected\n### second\n\nTUNING\n###\n",
    )

    selected = tuning.banner(banner_path, name="second")

    output = capsys.readouterr().out

    assert selected == expected
    assert "TUNING" in output


def test_banner_applies_custom_panel_styles_and_padding(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nSTYLED\n")

    selected = tuning.banner(
        banner_path,
        border_style="red",
        text_style="green",
        padding=(1, 2),
    )

    assert selected == "STYLED\n"
    assert len(console.renderables) == 1
    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.border_style == "red"
    assert panel.padding == (1, 2)
    assert isinstance(panel.renderable, Text)
    assert panel.renderable.style == "green"


def test_banner_applies_named_box_case_insensitively(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nBOXED\n")

    selected = tuning.banner(banner_path, box="rounded")

    assert selected == "BOXED\n"
    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.box is rich_box.ROUNDED


def test_banner_applies_mega_bold_box(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nMEGA\n")

    tuning.banner(banner_path, box="MEGA_BOLD")

    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.box is banners_module.MEGA_BOLD


def test_banner_accepts_box_object(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nOBJECT\n")
    custom_box = Box("┌─┬┐\n│ ││\n├─┼┤\n│ ││\n├─┼┤\n├─┼┤\n│ ││\n└─┴┘\n")

    tuning.banner(banner_path, box=custom_box)

    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.box is custom_box


def test_banner_applies_panel_background_style(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nBACKGROUND\n")

    tuning.banner(banner_path, background_style="on blue")

    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.style == "on blue"


def test_banner_defaults_to_independent_random_styles(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nRANDOM\n")
    choices = iter(["red", "blue"])
    monkeypatch.setattr(banners_module.random, "choice", lambda _: next(choices))

    tuning.banner(banner_path)

    panel = console.renderables[0]
    assert isinstance(panel, Panel)
    assert panel.border_style == "bold red"
    assert isinstance(panel.renderable, Text)
    assert panel.renderable.style == "bold blue"


def test_banner_can_render_without_border_but_with_padding(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nPLAIN\n")

    selected = tuning.banner(
        banner_path,
        border_style="red",
        text_style="cyan",
        padding=(1, 2, 3, 4),
        border=False,
    )

    assert selected == "PLAIN\n"
    assert len(console.renderables) == 1
    renderable = console.renderables[0]
    assert isinstance(renderable, Padding)
    assert (renderable.top, renderable.right, renderable.bottom, renderable.left) == (1, 2, 3, 4)
    assert isinstance(renderable.renderable, Text)
    assert renderable.renderable.style == "cyan"


def test_banner_applies_borderless_background_style(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    console = _record_console(monkeypatch)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nPLAIN\n")

    tuning.banner(banner_path, border=False, background_style="on magenta")

    renderable = console.renderables[0]
    assert isinstance(renderable, Padding)
    assert renderable.style == "on magenta"


def test_banner_accepts_directory_path(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 80)
    _write_banners(tmp_path / "banners.txt", "### only\nDIRECTORY\n")

    assert tuning.banner(tmp_path) == "DIRECTORY\n"


def test_banner_discovers_nearest_banners_file_from_calling_file(tmp_path: Path) -> None:
    repo_root = Path(__file__).parents[1]
    app_dir = tmp_path / "app"
    nested_dir = app_dir / "nested"
    nested_dir.mkdir(parents=True)
    _write_banners(app_dir / "banners.txt", "### only\nDISCOVERED\n")
    script_path = nested_dir / "run.py"
    script_path.write_text(
        "import sys\n"
        f"sys.path.insert(0, {str(repo_root)!r})\n"
        "import tuning\n"
        "selected = tuning.banner()\n"
        "print('RETURN=' + repr(selected))\n",
        encoding="utf8",
    )

    result = subprocess.run(
        [sys.executable, str(script_path)],
        check=True,
        capture_output=True,
        text=True,
    )

    assert "DISCOVERED" in result.stdout
    assert "RETURN='DISCOVERED\\n'" in result.stdout


def test_banner_discovers_from_cwd_when_caller_has_no_file(tmp_path: Path) -> None:
    repo_root = Path(__file__).parents[1]
    _write_banners(tmp_path / "banners.txt", "### only\nFROM CWD\n")

    result = subprocess.run(
        [
            sys.executable,
            "-c",
            "import sys; "
            f"sys.path.insert(0, {str(repo_root)!r}); "
            "import tuning; "
            "selected = tuning.banner(); "
            "print('RETURN=' + repr(selected))",
        ],
        cwd=tmp_path,
        check=True,
        capture_output=True,
        text=True,
    )

    assert "FROM CWD" in result.stdout
    assert "RETURN='FROM CWD\\n'" in result.stdout


def test_banner_randomly_selects_only_fitting_banners(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 20)
    banner_path = tmp_path / "banners.txt"
    _write_banners(
        banner_path,
        "### too_wide\nTHIS BANNER IS TOO WIDE\n### small\nSMALL\n",
    )

    def choice(options: Sequence[str]) -> str:
        assert options == ["SMALL\n"]
        return options[0]

    monkeypatch.setattr(banners_module.random, "choice", choice)

    assert tuning.banner(banner_path, border_style="red", text_style="blue") == "SMALL\n"


def test_banner_returns_none_when_no_banner_fits(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    _set_console_width(monkeypatch, 9)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nTOO WIDE\n")

    selected = tuning.banner(banner_path)
    output = capsys.readouterr().out

    assert selected is None
    assert "We" in output
    assert "tight" in output


def test_banner_named_too_wide_returns_none(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 9)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### wide\nTOO WIDE\n### small\nOK\n")

    assert tuning.banner(banner_path, name="wide") is None


def test_banner_padding_affects_terminal_fit(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 9)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nABCD\n")

    assert tuning.banner(banner_path, padding=(0, 2)) is None
    assert tuning.banner(banner_path, padding=(0, 1)) == "ABCD\n"


def test_banner_border_affects_terminal_fit(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 5)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nABCD\n")

    assert tuning.banner(banner_path) is None
    assert tuning.banner(banner_path, border=False) == "ABCD\n"


@pytest.mark.parametrize(
    "padding",
    [(), (1, 2, 3), (1, 2, 3, 4, 5), (-1,), ("1",)],
)
def test_banner_rejects_invalid_padding(tmp_path: Path, padding: tuple[object, ...]) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n")

    with pytest.raises(ValueError, match="padding"):
        tuning.banner(banner_path, padding=padding)  # type: ignore[arg-type]


def test_banner_rejects_non_boolean_border(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n")

    with pytest.raises(ValueError, match="border"):
        tuning.banner(banner_path, border="false")  # type: ignore[arg-type]


def test_banner_rejects_invalid_box_name(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n")

    with pytest.raises(ValueError, match="box"):
        tuning.banner(banner_path, box="not_a_box")


def test_banner_validates_box_when_border_is_false(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n")

    with pytest.raises(ValueError, match="box"):
        tuning.banner(banner_path, border=False, box="not_a_box")


def test_banner_raises_for_missing_discovered_banner_file(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.chdir(tmp_path)

    with pytest.raises(ValueError, match="Banner file not defined"):
        tuning.banner()


def test_banner_raises_key_error_for_missing_name(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n")

    with pytest.raises(KeyError):
        tuning.banner(banner_path, name="missing")


def test_banner_raises_for_invalid_banner_name(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### Not_Valid\nCONTENT\n")

    with pytest.raises(ValueError, match="Invalid banner name"):
        tuning.banner(banner_path)


def test_banner_raises_for_duplicate_banner_name(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### same\nONE\n### same\nTWO\n")

    with pytest.raises(ValueError, match="Duplicate banner name"):
        tuning.banner(banner_path)


def test_banner_allows_final_empty_marker_terminator(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_console_width(monkeypatch, 80)
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### only\nCONTENT\n###\n")

    assert tuning.banner(banner_path) == "CONTENT\n"


def test_banner_rejects_middle_empty_marker(tmp_path: Path) -> None:
    banner_path = tmp_path / "banners.txt"
    _write_banners(banner_path, "### first\nONE\n###\nTWO\n")

    with pytest.raises(ValueError, match="Empty banner marker"):
        tuning.banner(banner_path)



"""Public logging APIs for tuning."""

from __future__ import annotations

import logging
import logging.config
import logging.handlers
import sys
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, overload

from rich.cells import cell_len
from rich.console import Console
from rich.logging import RichHandler
from rich.text import Text

from tuning._config import (
    export_default_config,
    load_tuning_config,
    load_tuning_metadata,
    load_tuning_root_config,
    parse_size_to_bytes,
)
from tuning._levels import (
    get_level_spec,
    install_dynamic_level_methods,
    parse_runtime_level_spec,
    register_level_specs,
    validate_dynamic_level_methods,
    validate_level_specs,
)
from tuning._models import LevelSpec, PromptSpec
from tuning._prompt import render_prompt_text

ISO_FORMAT = "[%Y-%m-%d %H:%M:%S]"
"""ISO-like console timestamp format for `basicConfig(datefmt=...)`."""

DEFAULT_MAX_BYTES = parse_size_to_bytes("10 MB", field_name="DEFAULT_MAX_BYTES")
"""Default rotating file size used when only `backup_count` is provided."""

DEFAULT_BACKUP_COUNT = 3
"""Default number of rotated log files kept when only `max_bytes` is provided."""
_DETAILED_FILE_FORMAT = "%(asctime)s [%(levelname)s] %(name)s %(funcName)s: %(message)s"
_LEVEL_PREFIX_MIN_WIDTH = 3
_ZERO_CONFIG_LEVEL = "INFO"
_ROOT_PROMPT_SPEC = PromptSpec()
_DEFAULT_METADATA_INSTALLED = False
_ZERO_CONFIGURING = False


class TunedHandler(RichHandler):
    """Rich console handler that renders tuning level metadata.

    `TunedHandler` extends Rich's `RichHandler` with symbol/icon level
    prefixes, full-message level styling, and optional per-record boxes.
    It is normally created through `basicConfig()` or YAML configuration rather
    than instantiated directly.
    """

    def __init__(
        self,
        *args: Any,
        show_icon: bool = False,
        boxes: bool = False,
        **kwargs: Any,
    ) -> None:
        """Create a console handler.

        Args:
            *args: Positional arguments passed to `rich.logging.RichHandler`.
            show_icon: Use configured level icons instead of symbols when an
                icon is available.
            boxes: Render each console log record inside a Rich panel.
            **kwargs: Keyword arguments passed to `rich.logging.RichHandler`.
        """
        super().__init__(*args, **kwargs)
        self.show_icon = show_icon
        self.boxes = boxes

    def get_level_text(self, record: logging.LogRecord) -> Text:
        prefix = self._get_level_prefix(record)
        rendered_prefix = _pad_level_prefix(prefix)
        return Text.styled(rendered_prefix, _level_prefix_style(record))

    def _get_level_prefix(self, record: logging.LogRecord) -> str:
        spec = get_level_spec(record.levelname)
        if spec is None:
            return f"[{record.levelname}]"

        if self.show_icon and spec.icon:
            return spec.icon
        if spec.symbol:
            return spec.symbol
        return f"[{record.levelname}]"

    def render_message(self, record: logging.LogRecord, message: str) -> Any:
        message_renderable = super().render_message(record, message)
        style = _level_style(record)
        if style and isinstance(message_renderable, Text):
            message_renderable.stylize(style, 0, len(message_renderable))

        return message_renderable

    def render(
        self,
        *,
        record: logging.LogRecord,
        traceback: Any,
        message_renderable: Any,
    ) -> Any:
        from rich.containers import Renderables
        from rich.panel import Panel
        from rich.table import Table

        log_render = self._log_render
        output = Table.grid(padding=(0, 0))
        output.expand = True
        if log_render.show_time:
            output.add_column(style="log.time")
        if log_render.show_level and not self.boxes:
            output.add_column(style="log.level", width=log_render.level_width)
        output.add_column(ratio=1, style="log.message", overflow="fold")
        path = Path(record.pathname).name
        if log_render.show_path and path:
            output.add_column(style="log.path")

        row: list[Any] = []
        if log_render.show_time:
            row.append(self._render_log_time(record))
        if log_render.show_level and not self.boxes:
            row.append(self._render_level_with_separator(record))

        body = Renderables(
            [message_renderable] if not traceback else [message_renderable, traceback]
        )
        if self.boxes:
            row.append(
                Panel(
                    body,
                    title=self._render_box_title(record) if log_render.show_level else None,
                    title_align="left",
                    border_style=_level_prefix_style(record),
                    style=_level_prefix_style(record),
                    expand=True,
                )
            )
        else:
            row.append(body)

        if log_render.show_path and path:
            row.append(self._render_log_path(record, path))

        output.add_row(*row)
        return output

    def _render_box_title(self, record: logging.LogRecord) -> Text:
        title = Text(f"{self._get_level_prefix(record)} {record.levelname}")
        title.stylize(_level_prefix_style(record), 0, len(title))
        return title

    def _render_level_with_separator(self, record: logging.LogRecord) -> Text:
        level_text = self.get_level_text(record).copy()
        level_text.append(" ")
        level_text.stylize(_level_prefix_style(record), 0, len(level_text))
        return level_text

    def _render_log_time(self, record: logging.LogRecord) -> Text:
        log_render = self._log_render
        log_time = datetime.fromtimestamp(record.created)
        time_format: Any = self.formatter.datefmt if self.formatter is not None else None
        time_format = time_format or log_render.time_format

        if callable(time_format):
            log_time_display = time_format(log_time)
        else:
            log_time_display = Text(log_time.strftime(time_format))

        if log_time_display == log_render._last_time and log_render.omit_repeated_times:
            rendered_time = Text(" " * len(log_time_display))
        else:
            log_render._last_time = log_time_display
            rendered_time = log_time_display.copy()

        rendered_time.append(" ")
        return rendered_time

    def _render_log_path(self, record: logging.LogRecord, path: str) -> Text:
        link_path = record.pathname if self.enable_link_path else None
        path_text = Text(" ")
        path_text.append(path, style=f"link file://{link_path}" if link_path else "")
        if record.lineno:
            path_text.append(":")
            path_text.append(
                f"{record.lineno}",
                style=f"link file://{link_path}#{record.lineno}" if link_path else "",
            )
        return path_text


class TunedLogger(logging.Logger):
    """Logger subclass with YAML configuration and styled prompts.

    `TunedLogger` behaves like a normal stdlib logger while adding
    `from_yaml()` for named logger configuration and `prompt()` for styled
    interactive input.
    """

    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)
        self._tuning_signature: str | None = None
        self._prompt_spec: PromptSpec | None = None

    def isEnabledFor(self, level: int) -> bool:
        _ensure_zero_configured()
        return super().isEnabledFor(level)

    @classmethod
    def from_yaml(
        cls,
        config_path: str | Path,
        *,
        name: str = "app",
        defaults_path: str | Path | None = None,
        force: bool = False,
    ) -> TunedLogger:
        """Configure and return one named `TunedLogger` from YAML.

        The packaged `tuning/conf.yml` defaults are loaded first, then
        `config_path` is deep-merged on top. This method configures only the
        requested named logger. If the YAML defines only `root`, that section is
        used as the template for the named logger and does not configure the
        process root logger.

        Args:
            config_path: YAML override path.
            name: Logger name to configure.
            defaults_path: Optional defaults file used instead of the packaged
                config. Intended for tests and advanced integrations.
            force: Replace existing managed handlers for the same logger.

        Returns:
            The configured `TunedLogger`.

        Raises:
            ValueError: If the logger name is invalid, the logger already exists
                as a non-`TunedLogger`, or an existing configuration would be
                replaced without `force=True`.
        """
        if not isinstance(name, str) or not name.strip():
            raise ValueError("Logger name must be a non-empty string")

        resolved_config = load_tuning_config(
            config_path,
            logger_name=name,
            defaults_path=defaults_path,
        )
        logger = _get_or_create_tuned_logger(name)

        existing_signature = logger._tuning_signature
        if existing_signature == resolved_config.signature and not force:
            return logger

        if existing_signature and existing_signature != resolved_config.signature and not force:
            raise ValueError(
                f"TunedLogger {name!r} is already configured with a different config; use force=True"
            )

        if existing_signature is None and logger.handlers and not force:
            raise ValueError(
                f"TunedLogger {name!r} already has handlers but is not managed by from_yaml; use force=True"
            )

        register_level_specs(resolved_config.level_specs)
        install_dynamic_level_methods(cls, resolved_config.level_specs)

        if force:
            _close_logger_handlers(logger)

        logging.config.dictConfig(resolved_config.logging_config)
        configured_logger = logging.getLogger(name)
        if not isinstance(configured_logger, TunedLogger):
            raise ValueError(f"Configured logger {name!r} is not a TunedLogger")

        configured_logger._tuning_signature = resolved_config.signature
        configured_logger._prompt_spec = resolved_config.prompt_spec
        return configured_logger

    def prompt(
        self,
        message: str,
        *,
        password: bool = False,
        markup: bool | None = None,
    ) -> str:
        """Prompt for interactive input using configured prompt styling.

        Args:
            message: Prompt question to display.
            password: Hide input while typing.
            markup: Whether Rich markup in the prompt message is enabled. If
                omitted, the first inherited `TunedHandler` setting is used.

        Returns:
            The user's typed input.

        Raises:
            ValueError: If `message` is not a string.
            EOFError: If input cannot be read from the console.
        """
        if not isinstance(message, str):
            raise ValueError("Prompt message must be a string")

        _ensure_zero_configured()

        handler = _first_color_handler(self)
        console = handler.console if handler else Console()
        show_icon = handler.show_icon if handler else False
        use_markup = handler.markup if handler and markup is None else bool(markup)

        prompt_text = render_prompt_text(
            self._prompt_spec or _ROOT_PROMPT_SPEC,
            message,
            show_icon=show_icon,
            markup=use_markup,
        )
        return console.input(prompt_text, markup=False, password=password)


@overload
def getLogger(name: str) -> TunedLogger: ...


@overload
def getLogger(name: None = None) -> logging.Logger: ...


def getLogger(name: str | None = None) -> logging.Logger:
    """Return a logger using tuning's logger class for named loggers.

    Args:
        name: Logger name. If omitted, returns the process root logger.

    Returns:
        A `TunedLogger` for named loggers, or the stdlib root logger when
        `name` is omitted.

    Raises:
        ValueError: If `name` is empty or an existing logger with that name is
            not a `TunedLogger`.
    """
    _install_default_metadata()

    if name is None:
        return logging.getLogger()

    if not isinstance(name, str) or not name.strip():
        raise ValueError("Logger name must be a non-empty string")

    return _get_or_create_tuned_logger(name)


def export(path: str | Path | None = None, *, force: bool = False) -> Path:
    """Export the packaged default configuration to a YAML file.

    The exported file is copied from packaged `tuning/conf.yml`; it does not
    reconstruct the current runtime logging state.

    Args:
        path: Destination file path or existing directory. If omitted, writes
            `tuning.yml` next to the calling Python file.
        force: Overwrite an existing target file.

    Returns:
        The resolved path that was written.

    Raises:
        FileExistsError: If the target file exists and `force` is false.
        ValueError: If `force` is not a boolean.
    """
    _validate_bool("force", force)
    return export_default_config(_resolve_export_path(path), force=force)


def addLevel(
    num: int,
    name: str,
    symbol: str | None = None,
    icon: str | None = None,
    style: str | None = None,
) -> None:
    """Register a runtime-only custom log level and logger method.

    The level is added to stdlib logging for the current Python process and a
    matching method is installed on `TunedLogger`. For example,
    `addLevel(7, "MY_CUSTOM_LEVEL")` creates `logger.my_custom_level(...)` for
    existing and future tuning loggers. This does not update YAML files or
    exported configuration.

    Args:
        num: Numeric logging level code.
        name: Custom level name. Names are normalized to uppercase and must
            create a valid Python method name when lowercased.
        symbol: Compact console prefix used when icons are disabled.
        icon: Console prefix used when `show_icon=True`.
        style: Rich style applied to console level prefixes and messages.

    Raises:
        ValueError: If the level definition is invalid or conflicts with an
            existing logging level or dynamic method.
    """
    spec = parse_runtime_level_spec(
        num=num,
        name=name,
        symbol=symbol,
        icon=icon,
        style=style,
    )
    _install_default_metadata()
    validate_level_specs([spec])
    validate_dynamic_level_methods(TunedLogger, [spec])
    register_level_specs([spec])
    install_dynamic_level_methods(TunedLogger, [spec])


def _resolve_export_path(path: str | Path | None) -> Path:
    if path is None:
        return _caller_directory() / "tuning.yml"

    target_path = Path(path).expanduser()
    if target_path.is_dir():
        return target_path / "tuning.yml"

    return target_path


def _caller_directory() -> Path:
    try:
        caller_frame = sys._getframe(3)
    except ValueError:
        return Path.cwd()

    caller_file = caller_frame.f_globals.get("__file__")
    if not isinstance(caller_file, str):
        return Path.cwd()

    return Path(caller_file).resolve().parent


def basicConfig(
    *,
    filename: str | Path | None = None,
    filemode: str = "a",
    max_bytes: int | str | None = None,
    backup_count: int | None = None,
    format: str | None = None,
    datefmt: str | None = None,
    style: str = "%",
    level: int | str | None = logging.WARNING,
    stream: Any | None = None,
    handlers: list[logging.Handler] | None = None,
    force: bool = False,
    encoding: str | None = None,
    errors: str | None = None,
    console: bool = False,
    show_icon: bool = False,
    show_level: bool = True,
    show_path: bool = False,
    show_time: bool = False,
    boxes: bool = False,
    rich_tracebacks: bool = True,
    markup: bool = True,
    defaults_path: str | Path | None = None,
) -> None:
    """Configure the process root logger with tuning defaults.

    This follows stdlib `logging.basicConfig()` semantics: if the root logger
    already has handlers, the call is ignored unless `force=True` is passed.
    Without `filename`, a console `TunedHandler` is installed. With `filename`,
    a detailed file handler is installed; pass `console=True` to install both.

    Args:
        filename: File path for file logging. If omitted, console logging is
            configured instead.
        filemode: File open mode for file handlers.
        max_bytes: Rotation threshold in bytes or as a human-readable size
            string such as `"10 MB"`. Ignored when `filename` is omitted.
        backup_count: Number of rotated log files to keep. Ignored when
            `filename` is omitted.
        format: Formatter format string passed to `logging.basicConfig()`.
        datefmt: Formatter date format. For console timestamps, use this with
            `show_time=True`; `ISO_FORMAT` is provided as a convenience.
        style: Formatter style passed to `logging.basicConfig()`.
        level: Root logger level.
        stream: Console stream target.
        handlers: Explicit handlers. Cannot be combined with `stream`,
            `filename`, or `console`.
        force: Replace existing root handlers.
        encoding: File encoding for file handlers.
        errors: File encoding error handling for file handlers.
        console: Also create a console handler when `filename` is provided.
        show_icon: Use configured level icons instead of symbols.
        show_level: Show level prefixes or boxed level titles in console output.
        show_path: Show source file path in console output.
        show_time: Show timestamps in console output.
        boxes: Render each console log record inside a Rich panel.
        rich_tracebacks: Render Rich tracebacks in console output.
        markup: Enable Rich markup in console messages.
        defaults_path: Optional defaults file used instead of the packaged
            config. Intended for tests and advanced integrations.

    Raises:
        ValueError: If boolean options or rotation values are invalid, or if
            mutually exclusive handler options are combined.
    """
    root_logger = logging.getLogger()
    if root_logger.handlers and not force:
        return

    _validate_bool("console", console)
    _validate_bool("show_icon", show_icon)
    _validate_bool("show_level", show_level)
    _validate_bool("show_path", show_path)
    _validate_bool("show_time", show_time)
    _validate_bool("boxes", boxes)
    _validate_bool("rich_tracebacks", rich_tracebacks)
    _validate_bool("markup", markup)

    if handlers is not None and (filename is not None or stream is not None or console):
        raise ValueError(
            "'stream', 'filename', or 'console' should not be specified together with 'handlers'"
        )
    if filename is not None and stream is not None and not console:
        raise ValueError("'stream' and 'filename' should not be specified together")

    metadata = load_tuning_metadata(defaults_path=defaults_path)
    _install_tuning_metadata(metadata.level_specs, metadata.prompt_spec)

    logging_handlers = handlers
    if logging_handlers is None:
        logging_handlers = []
        if filename is not None:
            logging_handlers.append(
                _make_file_handler(
                    filename,
                    filemode=filemode,
                    max_bytes=max_bytes,
                    backup_count=backup_count,
                    encoding=encoding,
                    errors=errors,
                )
            )

        if filename is None or console:
            logging_handlers.append(
                _make_console_handler(
                    stream=stream,
                    show_icon=show_icon,
                    show_level=show_level,
                    show_path=show_path,
                    show_time=show_time,
                    boxes=boxes,
                    rich_tracebacks=rich_tracebacks,
                    markup=markup,
                )
            )

    basic_config_kwargs: dict[str, Any] = {
        "format": format,
        "datefmt": datefmt,
        "style": style,
        "level": level,
        "handlers": logging_handlers,
        "force": force,
    }
    logging.basicConfig(**basic_config_kwargs)


def _make_console_handler(
    *,
    stream: Any | None,
    show_icon: bool,
    show_level: bool,
    show_path: bool,
    show_time: bool,
    boxes: bool,
    rich_tracebacks: bool,
    markup: bool,
) -> TunedHandler:
    handler_kwargs: dict[str, Any] = {
        "show_icon": show_icon,
        "show_level": show_level,
        "show_path": show_path,
        "show_time": show_time,
        "boxes": boxes,
        "rich_tracebacks": rich_tracebacks,
        "markup": markup,
    }
    if stream is not None:
        handler_kwargs["console"] = Console(file=stream)

    return TunedHandler(**handler_kwargs)


def _make_file_handler(
    filename: str | Path,
    *,
    filemode: str,
    max_bytes: int | str | None,
    backup_count: int | None,
    encoding: str | None,
    errors: str | None,
) -> logging.FileHandler | logging.handlers.RotatingFileHandler:
    log_path = Path(filename).expanduser()
    log_path.parent.mkdir(parents=True, exist_ok=True)
    if max_bytes is None and backup_count is None:
        handler: logging.FileHandler | logging.handlers.RotatingFileHandler = logging.FileHandler(
            log_path,
            mode=filemode,
            encoding=encoding,
            errors=errors,
        )
    else:
        resolved_max_bytes = _resolve_max_bytes(max_bytes)
        resolved_backup_count = _resolve_backup_count(backup_count)
        handler = logging.handlers.RotatingFileHandler(
            log_path,
            mode=filemode,
            maxBytes=resolved_max_bytes,
            backupCount=resolved_backup_count,
            encoding=encoding,
            errors=errors,
        )
    handler.setFormatter(logging.Formatter(_DETAILED_FILE_FORMAT))
    return handler


def _resolve_max_bytes(max_bytes: int | str | None) -> int:
    if max_bytes is None:
        warnings.warn(
            "Rotation enabled, but no max_bytes specified. Using tuning.DEFAULT_MAX_BYTES.",
            stacklevel=4,
        )
        return DEFAULT_MAX_BYTES

    if isinstance(max_bytes, bool):
        raise ValueError("max_bytes must be a positive integer or size string")
    if isinstance(max_bytes, int):
        resolved_max_bytes = max_bytes
    elif isinstance(max_bytes, str):
        resolved_max_bytes = parse_size_to_bytes(max_bytes, field_name="max_bytes")
    else:
        raise ValueError("max_bytes must be a positive integer or size string")

    if resolved_max_bytes <= 0:
        raise ValueError("max_bytes must be greater than 0")

    return resolved_max_bytes


def _resolve_backup_count(backup_count: int | None) -> int:
    if backup_count is None:
        warnings.warn(
            "Rotation enabled, but no backup_count specified. Using tuning.DEFAULT_BACKUP_COUNT.",
            stacklevel=4,
        )
        return DEFAULT_BACKUP_COUNT

    if isinstance(backup_count, bool) or not isinstance(backup_count, int):
        raise ValueError("backup_count must be a non-negative integer")
    if backup_count < 0:
        raise ValueError("backup_count must be greater than or equal to 0")

    return backup_count


def _validate_bool(field_name: str, value: bool) -> None:
    if not isinstance(value, bool):
        raise ValueError(f"{field_name} must be a boolean")


def _level_style(record: logging.LogRecord) -> str | None:
    spec = get_level_spec(record.levelname)
    if spec is None:
        return None

    return spec.style


def _level_prefix_style(record: logging.LogRecord) -> str:
    spec = get_level_spec(record.levelname)
    if spec is not None and spec.style:
        return spec.style

    return f"logging.level.{record.levelname.lower()}"


def _pad_level_prefix(prefix: str) -> str:
    prefix_width = cell_len(prefix)
    if prefix_width >= _LEVEL_PREFIX_MIN_WIDTH:
        return prefix

    return prefix + " " * (_LEVEL_PREFIX_MIN_WIDTH - prefix_width)


def basicConfigFromYaml(
    config_path: str | Path | None = None,
    *,
    defaults_path: str | Path | None = None,
    force: bool = False,
) -> None:
    """Configure the process root logger from YAML.

    The packaged `tuning/conf.yml` defaults are loaded first. If `config_path`
    is provided, it is deep-merged on top. The resulting config is applied to
    the real process root logger with stdlib `logging.config.dictConfig()`.

    Args:
        config_path: Optional YAML override path. If omitted, packaged defaults
            are used directly.
        defaults_path: Optional defaults file used instead of the packaged
            config. Intended for tests and advanced integrations.
        force: Replace existing root handlers.

    Raises:
        ValueError: If the resolved YAML config is invalid.
    """
    root_logger = logging.getLogger()
    if root_logger.handlers and not force:
        return

    resolved_config = load_tuning_root_config(
        config_path,
        defaults_path=defaults_path,
    )
    _install_tuning_metadata(resolved_config.level_specs, resolved_config.prompt_spec)
    _prepare_configured_tuned_loggers(resolved_config.logging_config)

    if force:
        _close_logger_handlers(root_logger)

    logging.config.dictConfig(resolved_config.logging_config)


def _install_default_metadata() -> None:
    global _DEFAULT_METADATA_INSTALLED

    if _DEFAULT_METADATA_INSTALLED:
        return

    metadata = load_tuning_metadata()
    _install_tuning_metadata(metadata.level_specs, metadata.prompt_spec)
    _DEFAULT_METADATA_INSTALLED = True


def _ensure_zero_configured() -> None:
    global _ZERO_CONFIGURING

    if _ZERO_CONFIGURING:
        return

    _ZERO_CONFIGURING = True
    try:
        _install_default_metadata()
        root_logger = logging.getLogger()
        if root_logger.handlers:
            return

        zero_config_kwargs: dict[str, Any] = {
            "format": None,
            "level": _ZERO_CONFIG_LEVEL,
            "handlers": [
                _make_console_handler(
                    stream=None,
                    show_icon=False,
                    show_level=True,
                    show_path=False,
                    show_time=False,
                    boxes=False,
                    rich_tracebacks=True,
                    markup=True,
                )
            ],
        }
        logging.basicConfig(**zero_config_kwargs)
    finally:
        _ZERO_CONFIGURING = False


def _get_or_create_tuned_logger(name: str) -> TunedLogger:
    manager = logging.getLogger().manager
    existing = manager.loggerDict.get(name)

    if isinstance(existing, logging.Logger):
        logger = existing
    else:
        previous_logger_class = manager.loggerClass
        manager.loggerClass = TunedLogger
        try:
            logger = logging.getLogger(name)
        finally:
            manager.loggerClass = previous_logger_class

    if not isinstance(logger, TunedLogger):
        raise ValueError(
            f"Logger {name!r} already exists as {type(logger).__name__}; expected TunedLogger"
        )

    return logger


def _install_tuning_metadata(level_specs: list[LevelSpec], prompt_spec: PromptSpec) -> None:
    global _ROOT_PROMPT_SPEC

    register_level_specs(level_specs)
    install_dynamic_level_methods(TunedLogger, level_specs)
    _ROOT_PROMPT_SPEC = prompt_spec


def _prepare_configured_tuned_loggers(logging_config: dict[str, Any]) -> None:
    loggers = logging_config.get("loggers")
    if not isinstance(loggers, dict):
        return

    manager = logging.getLogger().manager
    for name in loggers:
        existing = manager.loggerDict.get(name)
        if existing is None or not isinstance(existing, logging.Logger):
            _get_or_create_tuned_logger(name)


def _close_logger_handlers(logger: logging.Logger) -> None:
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
        handler.close()


def _first_color_handler(logger: logging.Logger) -> TunedHandler | None:
    current: logging.Logger | None = logger
    while current is not None:
        for handler in current.handlers:
            if isinstance(handler, TunedHandler):
                return handler

        if not current.propagate:
            break
        current = current.parent

    return None


__all__ = [
    "TunedLogger",
    "TunedHandler",
    "LevelSpec",
    "PromptSpec",
    "ISO_FORMAT",
    "DEFAULT_MAX_BYTES",
    "DEFAULT_BACKUP_COUNT",
    "getLogger",
    "export",
    "addLevel",
    "basicConfig",
    "basicConfigFromYaml",
]
